跳到论文正文
W Papers独立论文阅读器 arXiv:2412.18008 论文正文 计算阅读时间 · 0%
引用这篇论文根据当前论文元数据生成,正式投稿前建议核对。

    
  
\setcctype

by

Parallel Contraction Hierarchies Can Be
Efficient and Scalable

Zijin Wan University of California, RiversideRiversideCAUSA [email protected] ,  Xiaojun Dong University of California, RiversideRiversideCAUSA [email protected] ,  Letong Wang University of California, RiversideRiversideCAUSA [email protected] ,  Enzuo Zhu University of California, DavisDavisCAUSA [email protected] ,  Yan Gu University of California, RiversideRiversideCAUSA [email protected]  and  Yihan Sun University of California, RiversideRiversideCAUSA [email protected]
(2025)
Abstract.

Contraction Hierarchies (CH) (Geisberger et al., 2008) is one of the most widely used algorithms for shortest-path queries on road networks. Compared to Dijkstra’s algorithm, CH enables orders of magnitude faster query performance through a preprocessing phase, which iteratively categorizes vertices into hierarchies and adds shortcuts. However, constructing a CH is an expensive task. Existing solutions, including parallel ones, may suffer from long construction time. Especially, in our experiments, we observe that existing parallel solutions demonstrate unsatisfactory scalability, and have performance close to sequential algorithms.

We present SPoCH (Scalable Parallelization of Contraction Hierarchies), an efficient and scalable CH construction algorithm in parallel. To address the challenges in previous work, our improvements focus on both redesigning the algorithm and leveraging parallel data structures. We compare SPoCH with the state-of-the-art sequential and parallel implementations on 16 graphs of various types. Our experiments show that SPoCH achieves 11111111–68×68\times68 × speedups over the best sequential baseline and 3.83.83.83.8–41×41\times41 × speedups over the best parallel baseline in CH construction, while maintaining competitive query performance and CH graph size. We released our code and all datasets used in this paper.

Parallel Algorithms, Graph Algorithms, Contraction Hierarchies, Shortest Paths
††journalyear: 2025††copyright: cc††conference: 2025 International Conference on Supercomputing; June 8–11, 2025; Salt Lake City, UT, USA††booktitle: 2025 International Conference on Supercomputing (ICS ’25), June 8–11, 2025, Salt Lake City, UT, USA††doi: 10.1145/3721145.3725744††isbn: 979-8-4007-1537-2/2025/06††ccs: Theory of computation Graph algorithms analysis††ccs: Theory of computation Shortest paths††ccs: Theory of computation Parallel algorithms††ccs: Theory of computation Shared memory algorithms

1. Introduction

Computing shortest distance is one of the most fundamental graph problems, playing a vital role in various applications, such as navigation on road networks. To accelerate point-to-point distance queries, many existing solutions use two-phase approaches, which preprocess the graph and construct an index—often an auxiliary graph—to facilitate queries. Among the most notable two-phase solutions is the Contraction Hierarchies (CH) (Geisberger et al., 2008), which is mainly designed for sparse networks such as road networks, and is widely used in practice, for instance in Google Maps. In addition to being used on its own, CH is also a vital component in other approaches for various applications on distance queries, such as Transit Node Routing (Arz et al., 2013; Bast et al., 2006, 2007), Hub-Based Labeling (Abraham et al., 2011, 2012; Delling et al., 2013a), and some renowned algorithms including CHASE (Bauer et al., 2010) and PHAST (Delling et al., 2013b). We refer the audience to the excellent surveys (Sommer, 2014; Bast et al., 2016; Madkour et al., 2017) for more background of the state-of-the-art techniques for route planning.

Algorithms Preprocessing time Query time
Dijkstra (Dijkstra, 1959) - 7.20 s
ΔΔ\Deltaroman_Δ-stepping∗ (Dong et al., 2024b) - 0.320 s
RoutingKit (contributors, 2024; Dibbelt et al., 2016) 2466 s 79.1 μ𝜇\muitalic_μs
PHAST (Group, 2024; Delling et al., 2013b) 1341 s 138 μ𝜇\muitalic_μs
CH-Constructor∗ (Bühler and Nusser, 2024) 1527 s 317 μ𝜇\muitalic_μs
OSRM∗ (Luxen and Vetter, 2024, 2011) 307307307307s 163 μ𝜇\muitalic_μs
Ours∗ 23.1 s 93.3 μ𝜇\muitalic_μs
Table 1. Preprocessing and point-to-point shortest path query time on the road network North America (OpenStreetMap contributors, 2010). The graph has 87 million vertices and 113 million edges. 1 μ𝜇\muitalic_μs = 10−6superscript10610^{-6}10 start_POSTSUPERSCRIPT - 6 end_POSTSUPERSCRIPT s. (*) denotes parallel implementations; others are sequential. The reported query time is the average across 1000100010001000 randomly selected pairs. Dijkstra (Dijkstra, 1959) and ΔΔ\Deltaroman_Δ-stepping (Dong et al., 2024b) are sequential and parallel SSSP algorithms for comparison.

In this paper, we consider a graph G=(V,E)𝐺𝑉𝐸G=(V,E)italic_G = ( italic_V , italic_E ) with an edge weight function w:E↦ℝ+:𝑤maps-to𝐸superscriptℝw:E\mapsto\mathbb{R}^{+}italic_w : italic_E ↦ blackboard_R start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT. Given two vertices s,t∈V𝑠𝑡𝑉s,t\in Vitalic_s , italic_t ∈ italic_V, a distance query asks for the shortest distance from s𝑠sitalic_s to t𝑡titalic_t. The idea of CH, as illustrated in Fig. 1(a), is to “contract” the graph into a hierarchy, which is an auxiliary graph GCHsubscript𝐺CH\mathit{G_{\scriptsize\mbox{\it CH}}}italic_G start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT that preserves the shortest distances in G𝐺Gitalic_G, such that distance queries on GCHsubscript𝐺CH\mathit{G_{\scriptsize\mbox{\it CH}}}italic_G start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT can be much faster. In each iteration, one vertex u𝑢uitalic_u and all its edges are moved from the original graph to an auxiliary graph GCHsubscript𝐺CH\mathit{G_{\scriptsize\mbox{\it CH}}}italic_G start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT, forming a level in GCHsubscript𝐺CH\mathit{G_{\scriptsize\mbox{\it CH}}}italic_G start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT. Additionally, extra edges (shortcuts) are added to the original graph, forming the overlay graph GOsubscript𝐺O{\mathit{G_{\scriptsize\mbox{\it O}}}}italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT, to preserve the pairwise distances among the remaining vertices. Finally, all vertices in GCHsubscript𝐺CH\mathit{G_{\scriptsize\mbox{\it CH}}}italic_G start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT form the contraction hierarchies (CH). A distance query requires a bidirectional search on GCHsubscript𝐺CH\mathit{G_{\scriptsize\mbox{\it CH}}}italic_G start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT.

Empirically, a distance query on the CH touches far fewer vertices than in the original graph, thereby resulting in much faster query speed. As shown in Tab. 1, for the North American roadmap from OpenStreetMap (OpenStreetMap contributors, 2010), queries can be 103superscript10310^{3}10 start_POSTSUPERSCRIPT 3 end_POSTSUPERSCRIPT–105superscript10510^{5}10 start_POSTSUPERSCRIPT 5 end_POSTSUPERSCRIPT faster than directly using SSSP algorithms, such as Dijkstra (Dijkstra, 1959) and ΔΔ\Deltaroman_Δ-stepping (a parallel SSSP algorithm) (Meyer and Sanders, 2003; Dong et al., 2021). However, this impressive querying speed comes at a cost—constructing GCHsubscript𝐺CH\mathit{G_{\scriptsize\mbox{\it CH}}}italic_G start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT is expensive. Existing sequential solutions use 22–41 minutes, and the best parallel solutions need more than 300 seconds. Such long preprocessing time may require significant computational resources, and limit its adaptability to large graphs. Such running time is reasonable for a sequential algorithm—both RoutingKit (Dibbelt et al., 2016) and PHAST (Delling et al., 2013b) take 100×100\times100 × construction time than Dijkstra, which is acceptable considering the speedup for queries. However, the existing parallel CH algorithms are only 4.4×4.4\times4.4 × faster than (sequential) PHAST even on a 96-core machine. The results indicate a significant gap, suggesting the potential for improvements in parallelism in CH construction algorithms.

In this paper, we propose SPoCH (Scalable Parallelization of Contraction Hierarchies), which supports scalable and efficient contraction hierarchies construction with high parallelism without compromising the query performance. The performance gain is mainly from good parallelism and algorithmic improvements for sequential performance.

Conceptually, parallelizing CH construction is not hard. In 2009, Vetter (Vetter, 2009) pointed out that vertices in an independent set (i.e., vertices that do not share edges) can be contracted in parallel, as illustrated in Fig. 1(b). However, despite being studied in many later papers (Luxen and Vetter, 2024, 2011; Karimi et al., 2019, 2020; Kieritz et al., 2010; Bühler and Nusser, 2024), we are unaware of scalable multicore CH implementations. In fact, constructing CH is a sophisticated process with intensive computation. Achieving good performance requires high parallelism in all steps and careful algorithmic design. We observe two major challenges. First, identifying the vertices to contract involves simulating contractions on many (if not most) vertices, which calculates distances for numerous vertex pairs. Such a process is expensive, requiring high parallelism to enable good performance. The second challenge is to dynamically maintain the graph in parallel, since the graph may experience rapid changes, such as the removal of contracted vertices/edges and the insertion of shortcuts. Existing solutions may sacrifice parallelism (e.g., using locks) to support such dynamic updates. Ideally, a scalable solution would necessitate graph maintenance support with full parallelism.

Contributions.  We propose SPoCH, which consists of a new algorithmic framework for CH, and a high-performance implementation. Our improvement lies in both algorithm and data structure design with high parallelism. Algorithmically, we introduce the new LocalSearch step in construction to address the high cost of simulated contraction on a large fraction of vertices. This step gathers all the pairs of vertices for distance computation, and processes them in a batch in parallel. Processing them as a batch allows for combining and eliminating repeated computations, offering high potential to exploit parallelism, as well as a pruning process to remove suboptimal shortcuts added in previous rounds. Finally, SPoCH memoizes the computed distances in this step to facilitate later computation. We provide more information in Sec. 3.1. Combining these benefits, the LocalSearch step improves both the sequential running time and parallelism.

Another key performance improvement in SPoCH comes from using efficient parallel data structures. As mentioned, to get a highly parallel solution, we need to efficiently handle edge updates to the graph in parallel. We use the phase-concurrent hash table (Shun and Blelloch, 2014) to buffer newly added shortcuts in each round, which allows for efficient lock-free concurrent insertions of new edges. However, combining these edges to the overlay graph in each round requires scanning the entire hash table array and the edge list of the overlay graph. This may be inefficient in most rounds where only a small number of shortcuts are added. Our solution employs a lazy update scheme to delay the combination of the shortcuts to the overlay graph, while still making the delayed shortcuts visible to future computations. We introduce this technique in Sec. 4. This approach reduces the cost of maintaining the graph from 39.9%percent39.939.9\%39.9 % to 19.8%percent19.819.8\%19.8 % of the overall running time on average across all tested graphs.

Refer to caption
Figure 1. Illustration of the construction of Contraction Hierarchies. For simplicity, we assume the input graph has unit weights, omitting the weight “1” from the graph.

We compare SPoCH with the state-of-the-art sequential and parallel solutions on various graphs. Even the sequential running time of SPoCH is as fast as the sequential baselines. Due to our new design, SPoCH achieves high parallelism and good scalability with large numbers of processors (see Fig. 6). On North America in Tab. 1, SPoCH is 13.313.313.313.3–107×107\times107 × faster in construction than the baselines while achieving similar query performance. On the 16 graphs we tested in Tab. 4, compared to the fastest baseline on each graph, SPoCH is 3.83–41.0×\times× faster, with an average of 8.85×8.85\times8.85 ×. We also conduct in-depth performance studies of the CH algorithms in Sec. 6. We released our code in (Wan et al., 2024).

Preliminaries.  We focus on the shared-memory setting with fork-join parallelism. The computation starts with one thread. A thread can fork two child software threads to work in parallel. When both children complete, the parent thread continues. Such a computation can be executed by a randomized work-stealing scheduler (Blumofe and Leiserson, 1999; Arora et al., 2001; Blelloch et al., 2020b). We use the atomic operation compare_and_swap(p,vold,vnew)𝑝subscript𝑣oldsubscript𝑣new(p,v_{{\scriptsize\mbox{\it old}}},v_{{\scriptsize\mbox{\it new}}})( italic_p , italic_v start_POSTSUBSCRIPT old end_POSTSUBSCRIPT , italic_v start_POSTSUBSCRIPT new end_POSTSUBSCRIPT ), which checks if the memory location pointed to by p𝑝pitalic_p has the value voldsubscript𝑣oldv_{\scriptsize\mbox{\it old}}italic_v start_POSTSUBSCRIPT old end_POSTSUBSCRIPT, and if so, changes the value to vnewsubscript𝑣newv_{\scriptsize\mbox{\it new}}italic_v start_POSTSUBSCRIPT new end_POSTSUBSCRIPT. The function returns true if it successfully changes the value, and false otherwise. We also use write_min(p,v)𝑝𝑣(p,\mathit{v})( italic_p , italic_v ), which reads the memory location pointed to by p𝑝pitalic_p and writes value v𝑣vitalic_v to it if v𝑣vitalic_v is smaller than the current value.

We consider a weighted graph G=(V,E,w)𝐺𝑉𝐸𝑤G=(V,E,w)italic_G = ( italic_V , italic_E , italic_w ) with n=|V|𝑛𝑉n=|V|italic_n = | italic_V |, m=|E|𝑚𝐸m=|E|italic_m = | italic_E |, and an edge weight function w:E→ℝ+:𝑤→𝐸superscriptℝw:E\to\mathbb{R}^{+}italic_w : italic_E → blackboard_R start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT. For v∈V𝑣𝑉v\in Vitalic_v ∈ italic_V, we define Nin⁢(v)={u|(u,v)∈E}subscript𝑁in𝑣conditional-set𝑢𝑢𝑣𝐸N_{\scriptsize\mbox{\it in}}(v)=\{u\,|\,(u,v)\in E\}italic_N start_POSTSUBSCRIPT in end_POSTSUBSCRIPT ( italic_v ) = { italic_u | ( italic_u , italic_v ) ∈ italic_E } as the in-neighbors of v𝑣vitalic_v, and Nout⁢(v)={u|(v,u)∈E}subscript𝑁out𝑣conditional-set𝑢𝑣𝑢𝐸N_{\scriptsize\mbox{\it out}}(v)=\{u\,|\,(v,u)\in E\}italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_v ) = { italic_u | ( italic_v , italic_u ) ∈ italic_E } as the out-neighbors of v𝑣vitalic_v. We use N⁢(v)=Nin⁢(v)∪Nout⁢(v)𝑁𝑣subscript𝑁in𝑣subscript𝑁out𝑣N(v)=N_{\scriptsize\mbox{\it in}}(v)\cup N_{\scriptsize\mbox{\it out}}(v)italic_N ( italic_v ) = italic_N start_POSTSUBSCRIPT in end_POSTSUBSCRIPT ( italic_v ) ∪ italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_v ). For a set of vertices V′⊂Vsuperscript𝑉′𝑉V^{\prime}\subset Vitalic_V start_POSTSUPERSCRIPT ′ end_POSTSUPERSCRIPT ⊂ italic_V, we define N⁢(V′)𝑁superscript𝑉′N(V^{\prime})italic_N ( italic_V start_POSTSUPERSCRIPT ′ end_POSTSUPERSCRIPT ) as all neighbors of vertices in V′superscript𝑉′V^{\prime}italic_V start_POSTSUPERSCRIPT ′ end_POSTSUPERSCRIPT. Note that the main process of the CH construction algorithm is to update the overlay graph GOsubscript𝐺𝑂G_{O}italic_G start_POSTSUBSCRIPT italic_O end_POSTSUBSCRIPT by removing contracted vertices and incident edges, and inserting shortcuts. Therefore, when we use Nin⁢(⋅),Nout⁢(⋅),N⁢(⋅)subscript𝑁in⋅subscript𝑁out⋅𝑁⋅N_{\scriptsize\mbox{\it in}}(\cdot),N_{\scriptsize\mbox{\it out}}(\cdot),N(\cdot)italic_N start_POSTSUBSCRIPT in end_POSTSUBSCRIPT ( ⋅ ) , italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( ⋅ ) , italic_N ( ⋅ ) in the algorithm description, we always refer to the neighborhood of a vertex on the overlay graph.

G=(V,E)𝐺𝑉𝐸G=(V,E)italic_G = ( italic_V , italic_E ) : The input graph
GO=(VO,EO)subscript𝐺Osubscript𝑉Osubscript𝐸O{{\mathit{G_{\scriptsize\mbox{\it O}}}}}=(\mathit{V_{\scriptsize\mbox{\it O}}}% ,{\mathit{E_{\scriptsize\mbox{\it O}}}})italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT = ( italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT , italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT ) : The overlay graph
w⁢(u,v)𝑤𝑢𝑣{w(u,v)}italic_w ( italic_u , italic_v ) : Weight for edge (u,v)𝑢𝑣(u,v)( italic_u , italic_v )
P⁢[v]𝑃delimited-[]𝑣{\mathit{P}[v]}italic_P [ italic_v ] : The score of vertex v𝑣vitalic_v
Nin⁢(v)subscript𝑁in𝑣{N_{\scriptsize\mbox{\it in}}(v)}italic_N start_POSTSUBSCRIPT in end_POSTSUBSCRIPT ( italic_v ) : In-neighbors of vertex v𝑣vitalic_v
Nout⁢(v)subscript𝑁out𝑣{N_{\scriptsize\mbox{\it out}}(v)}italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_v ) : Out-neighbors of vertex v𝑣vitalic_v
N⁢(v)𝑁𝑣{N(v)}italic_N ( italic_v ) : In- and out-neighbors of vertex v𝑣vitalic_v
(Unless otherwise specified, we use Nin⁢(v)subscript𝑁in𝑣N_{\scriptsize\mbox{\it in}}(v)italic_N start_POSTSUBSCRIPT in end_POSTSUBSCRIPT ( italic_v ), Nout⁢(v)subscript𝑁out𝑣N_{\scriptsize\mbox{\it out}}(v)italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_v ) and N⁢(v)𝑁𝑣N(v)italic_N ( italic_v ) to refer
to the neighbors of v𝑣vitalic_v on the overlay graph)
Table 2. General notations used for contraction hierarchies.

2. Contraction Hierarchies

Contraction Hierarchies (CH) was proposed by Geisberger et al. (Geisberger et al., 2008) in 2008, building upon the concept of highway hierarchies (Sanders and Schultes, 2006; Knopp et al., 2007). CH is one of the most widely adopted techniques in route planning and is used in many systems, such as Google Maps.

Let G=(V,E)𝐺𝑉𝐸G=(V,E)italic_G = ( italic_V , italic_E ) be the input graph. Its Contraction Hierarchies 𝑮CH=(𝑽,𝑬CH)subscript𝑮CH𝑽subscript𝑬CH\boldsymbol{\mathit{G_{\scriptsize\mbox{\it CH}}}=(V,\mathit{\mathit{E_{% \scriptsize\mbox{\it CH}}}})}bold_italic_G start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT bold_= bold_( bold_italic_V bold_, bold_italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT bold_) is an auxiliary graph that preserves the pairwise distances of G𝐺Gitalic_G. Distance queries on GCHsubscript𝐺CH\mathit{G_{\scriptsize\mbox{\it CH}}}italic_G start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT can be orders of magnitude faster than on G𝐺Gitalic_G itself, particularly for sparse graphs such as road networks. The CH algorithm iteratively contracts the “least important” vertex (determined by a user-specified score function) from the graph, and adds its edges to the contraction hierarchy. Additional edges (shortcuts) are added to the remaining graph (referred to as the overlay graph GOsubscript𝐺O{\mathit{G_{\scriptsize\mbox{\it O}}}}italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT) to preserve the distances. Each contracted vertex represents a level in the CH, forming a hierarchy. This process is illustrated in Fig. 1(a). A distance query from s𝑠sitalic_s to t𝑡titalic_t is performed by running a bidirectional search from both s𝑠sitalic_s and t𝑡titalic_t on the CH. The query algorithm is presented in Sec. 5. By contracting vertices and adding shortcuts, the hop distances between vertices are greatly reduced, allowing for much faster query performance.

Given its wide range of applications (Funke et al., 2019, 2015; Abeywickrama et al., 2016; Sanchez et al., 2022; Ouyang et al., 2020), CH has been extensively studied. In this section, we first review the algorithm proposed by Geisberger et al. (Geisberger et al., 2008) along with the fundamental concepts. Then in Sec. 2.2, we introduce existing parallel solutions. Other related studies are discussed in Sec. 7.

Input: A graph G=(V,E)𝐺𝑉𝐸G=(V,E)italic_G = ( italic_V , italic_E ).
Output: The contraction hierarchy GCH=(V,ECH)subscript𝐺CH𝑉subscript𝐸CH\mathit{G_{\scriptsize\mbox{\it CH}}}=(V,\mathit{\mathit{E_{\scriptsize\mbox{% \it CH}}}})italic_G start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT = ( italic_V , italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT ).
Maintains : The overlay graph GO=(VO,EO)subscript𝐺Osubscript𝑉Osubscript𝐸O{\mathit{G_{\scriptsize\mbox{\it O}}}}=(\mathit{V_{\scriptsize\mbox{\it O}}},{% \mathit{E_{\scriptsize\mbox{\it O}}}})italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT = ( italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT , italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT )
The score of each vertex P⁢[⋅]𝑃delimited-[]⋅\mathit{P}[\cdot]italic_P [ ⋅ ]
1
2GO←G,ECH←∅formulae-sequence←subscript𝐺O𝐺←subscript𝐸CH{\mathit{G_{\scriptsize\mbox{\it O}}}}\leftarrow G,\mathit{\mathit{E_{% \scriptsize\mbox{\it CH}}}}\leftarrow\emptysetitalic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT ← italic_G , italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT ← ∅
3 Compute the score P⁢[⋅]𝑃delimited-[]⋅\mathit{P}[\cdot]italic_P [ ⋅ ] for all v∈V𝑣𝑉v\in Vitalic_v ∈ italic_V
4 while VO≠∅subscript𝑉O\mathit{V_{\scriptsize\mbox{\it O}}}\neq\emptysetitalic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT ≠ ∅ do
       // Select the vertex v𝑣vitalic_v with lowest score from VOsubscript𝑉O\mathit{V_{\scriptsize\mbox{\it O}}}italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT
5       v=arg⁡minv′∈VO⁡P⁢[v′]𝑣subscriptsuperscript𝑣′subscript𝑉O𝑃delimited-[]superscript𝑣′v=\operatorname*{\arg\!\min}_{v^{\prime}\in\mathit{V_{\scriptsize\mbox{\it O}}% }}\mathit{P}[v^{\prime}]italic_v = start_OPERATOR roman_arg roman_min end_OPERATOR start_POSTSUBSCRIPT italic_v start_POSTSUPERSCRIPT ′ end_POSTSUPERSCRIPT ∈ italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT end_POSTSUBSCRIPT italic_P [ italic_v start_POSTSUPERSCRIPT ′ end_POSTSUPERSCRIPT ] for u1∈Nin⁢(v)subscript𝑢1subscript𝑁in𝑣u_{1}\in N_{\scriptsize\mbox{\it in}}(v)italic_u start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT ∈ italic_N start_POSTSUBSCRIPT in end_POSTSUBSCRIPT ( italic_v ) do
6             for u2∈Nout⁢(v)subscript𝑢2subscript𝑁out𝑣u_{2}\in N_{\scriptsize\mbox{\it out}}(v)italic_u start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ∈ italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_v ) do
                   // Shortest path distance from u1subscript𝑢1u_{1}italic_u start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT to u2subscript𝑢2u_{2}italic_u start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT
7                   l←WPS⁢(u1,u2)←𝑙WPSsubscript𝑢1subscript𝑢2l\leftarrow\textsc{WPS}(u_{1},u_{2})italic_l ← WPS ( italic_u start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_u start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT )
8                   if w⁢(u1,v)+w⁢(v,u2)=l𝑤subscript𝑢1𝑣𝑤𝑣subscript𝑢2𝑙w(u_{1},v)+w(v,u_{2})=litalic_w ( italic_u start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_v ) + italic_w ( italic_v , italic_u start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ) = italic_l then
9                         EO=EO∪{(u1,u2)}subscript𝐸Osubscript𝐸Osubscript𝑢1subscript𝑢2{\mathit{E_{\scriptsize\mbox{\it O}}}}={\mathit{E_{\scriptsize\mbox{\it O}}}}% \cup\{(u_{1},u_{2})\}italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT = italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT ∪ { ( italic_u start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_u start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ) }
10                         w⁢(u1,u2)=w⁢(u1,v)+w⁢(v,u2)𝑤subscript𝑢1subscript𝑢2𝑤subscript𝑢1𝑣𝑤𝑣subscript𝑢2w(u_{1},u_{2})=w(u_{1},v)+w(v,u_{2})italic_w ( italic_u start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_u start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ) = italic_w ( italic_u start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_v ) + italic_w ( italic_v , italic_u start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT )
11      VA←N⁢(v)←subscript𝑉𝐴𝑁𝑣\mathit{V_{A}}\leftarrow N(v)italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT ← italic_N ( italic_v )
12       Remove v𝑣vitalic_v and all its edges from GO=(VO,EO)subscript𝐺Osubscript𝑉Osubscript𝐸O{\mathit{G_{\scriptsize\mbox{\it O}}}}=(\mathit{V_{\scriptsize\mbox{\it O}}},{% \mathit{E_{\scriptsize\mbox{\it O}}}})italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT = ( italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT , italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT ) and add them to GCHsubscript𝐺CH\mathit{G_{\scriptsize\mbox{\it CH}}}italic_G start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT
13       Update the scores P⁢[u]𝑃delimited-[]𝑢\mathit{P}[u]italic_P [ italic_u ] for all u∈VA𝑢subscript𝑉𝐴u\in\mathit{V_{A}}italic_u ∈ italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT
return GCHsubscript𝐺CH\mathit{G_{\scriptsize\mbox{\it CH}}}italic_G start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT
Algorithm 1 Contraction Hierarchies Construction (Geisberger et al., 2008)

2.1. Sequential Solutions

We present the sequential construction algorithm for CH in Alg. 1. The algorithm takes a graph G=(V,E)𝐺𝑉𝐸G=(V,E)italic_G = ( italic_V , italic_E ) as the input, and computes the CH for G𝐺Gitalic_G in GCH=(V,ECH)subscript𝐺CH𝑉subscript𝐸CH\mathit{G_{\scriptsize\mbox{\it CH}}}=(V,\mathit{\mathit{E_{\scriptsize\mbox{% \it CH}}}})italic_G start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT = ( italic_V , italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT ). As mentioned, the contraction process requires removing (contracting) vertices, and adding shortcuts to preserve the shortest distances. To avoid destroying the input graph, an overlay graph GO=(VO,EO)subscript𝐺Osubscript𝑉Osubscript𝐸O{\mathit{G_{\scriptsize\mbox{\it O}}}}=(\mathit{V_{\scriptsize\mbox{\it O}}},{% \mathit{E_{\scriptsize\mbox{\it O}}}})italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT = ( italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT , italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT ) is used to reflect the changes. A score array P⁢[⋅]𝑃delimited-[]⋅\mathit{P}[\cdot]italic_P [ ⋅ ] is also maintained for all uncontracted vertices, representing their “importance” and deciding the contracting order.

The algorithm starts by computing the initial scores of all vertices (line 1). Then, it iteratively selects and contracts the vertex with the lowest score (line 1) until the overlay graph is empty (line 1). When contracting a vertex v𝑣vitalic_v, the algorithm computes the shortest paths (line 1) from each u1∈Nin⁢(v)subscript𝑢1subscript𝑁in𝑣u_{1}\in N_{\scriptsize\mbox{\it in}}(v)italic_u start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT ∈ italic_N start_POSTSUBSCRIPT in end_POSTSUBSCRIPT ( italic_v ) to each u2∈Nout⁢(v)subscript𝑢2subscript𝑁out𝑣u_{2}\in N_{\scriptsize\mbox{\it out}}(v)italic_u start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ∈ italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_v ) by running Dijkstra’s algorithm. These shortest path queries are referred to as the Witness Path Search (WPS) (Geisberger et al., 2008; Karimi et al., 2019). If the shortest path length l𝑙litalic_l from u1subscript𝑢1u_{1}italic_u start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT to u2subscript𝑢2u_{2}italic_u start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT is the same as w⁢(u1,v)+w⁢(v,u2)𝑤subscript𝑢1𝑣𝑤𝑣subscript𝑢2w(u_{1},v)+w(v,u_{2})italic_w ( italic_u start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_v ) + italic_w ( italic_v , italic_u start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ) (line 1), then v𝑣vitalic_v can be on the shortest path from u1subscript𝑢1u_{1}italic_u start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT to u2subscript𝑢2u_{2}italic_u start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT. Thus, the shortcut edge (u1,u2)subscript𝑢1subscript𝑢2(u_{1},u_{2})( italic_u start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_u start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ) with weight l𝑙litalic_l is added to EOsubscript𝐸O{\mathit{E_{\scriptsize\mbox{\it O}}}}italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT to preserve the distances on EOsubscript𝐸O{\mathit{E_{\scriptsize\mbox{\it O}}}}italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT (line 1 and 1). After that, v𝑣vitalic_v and all its incident edges are removed from the overlay graph GO=(VO,EO)subscript𝐺Osubscript𝑉Osubscript𝐸O{\mathit{G_{\scriptsize\mbox{\it O}}}}=(\mathit{V_{\scriptsize\mbox{\it O}}},{% \mathit{E_{\scriptsize\mbox{\it O}}}})italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT = ( italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT , italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT ), and are added to the CH GCHsubscript𝐺CH\mathit{G_{\scriptsize\mbox{\it CH}}}italic_G start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT. Contracting v𝑣vitalic_v may cause the score of all its neighbors to change, which are updated on line 1. Finally, the algorithm moves on to contract the next vertex, until all vertices are contracted.

Vertex Score and Edge Difference.  Scoring the vertices is the most time-consuming part of CH construction and significantly impacts the shortest path query performance in CH. The most widely used score function is the edge difference (Geisberger et al., 2008). For a vertex v𝑣vitalic_v, the edge difference is the change in the number of edges on the overlay graph GOsubscript𝐺O{\mathit{G_{\scriptsize\mbox{\it O}}}}italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT after contracting v𝑣vitalic_v, i.e., the number of shortcuts added minus the number of edges incident to v𝑣vitalic_v. A vertex with a higher edge difference is considered more important because its contraction has a larger impact on the graph’s structure. To compute the number of shortcuts to be added, a simulated contraction on v𝑣vitalic_v is required, which virtually generates and counts the shortcuts. This simulation is very similar to line 1–1 except it does not add edges or modify edge weights in GOsubscript𝐺O{\mathit{G_{\scriptsize\mbox{\it O}}}}italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT.

In existing work (Geisberger et al., 2008, 2012), the score function is usually defined as a weighted combination of edge difference and many other metrics, such as vertex degree, hop distance, and hierarchy depth. Among them, edge difference has the highest weight and is usually the most computationally expensive part. Therefore, for simplicity, our algorithm description uses edge difference as the score function. Our algorithmic ideas are independent of the score function.

2.2. Existing Parallel Algorithms

The construction for CH is very costly, so it is natural to consider parallelizing this process by contracting many vertices in parallel. However, note that we cannot contract arbitrary vertices in parallel. For instance, if two adjacent vertices u𝑢uitalic_u and v𝑣vitalic_v are contracted together, u𝑢uitalic_u may choose to add a shortcut incident to v𝑣vitalic_v. Since v𝑣vitalic_v is removed from the overlay graph at the same time, the shortcut may not be added correctly, resulting in distances not being preserved accurately on the overlay graph. In 2009, Vetter (Vetter, 2009) first observed that vertices that do not share edges (i.e., an independent set) can be contracted in parallel, and proposed the first parallel CH construction algorithm. In this case, all added shortcuts will be connected to uncontracted vertices and are thus safe to add to the overlay graph.

Almost all later parallel solutions (e.g., (Karimi et al., 2020; Bühler and Nusser, 2024; Luxen and Vetter, 2011; Kieritz et al., 2010; Karimi et al., 2019)) follow Vetter’s high-level idea. In each round, an independent set (IS) of vertices is identified to be contracted. All these vertices will be placed on the same level in the CH. After that, shortcuts will be added normally. This process is repeated until all vertices have been contracted. As with the sequential algorithm that first contracts vertices with the lowest score, the vertices in the IS should also have low scores in general. In Vetter’s original algorithm, the IS includes all vertices with the smallest score within its k𝑘kitalic_k-hop neighborhood. OSRM (Luxen and Vetter, 2024, 2011) specifically uses k=2𝑘2k=2italic_k = 2. Our implementation also follows this idea and uses k=1𝑘1k=1italic_k = 1. An existing GPU algorithm (Karimi et al., 2019) and CH-Constructor (Bühler and Nusser, 2024) find a maximal independent set (MIS) and include all vertices in this MIS with scores smaller than a threshold.

Challenges to Achieve High-Performance Implementations.  While Vetter’s work reveals the potential parallelism in the algorithm, the idea itself is not sufficient to enable a high-performance solution. Many challenges remain in other parts of the algorithm. We highlight two components here. The first part is the process to (re)compute the scores of vertices. When multiple vertices are contracted together, a large number of simulated contractions is needed, which all involve running WPSes by Dijkstra’s algorithm. The second challenge is to update the overlay graph in parallel, since contracting vertices in parallel results in a bulk of new shortcut edges that need to be added to the overlay graph. Both parts (performing WPS and updating the graph) involve expensive computation, and require careful algorithm design and parallelization. In our experiments, we test two existing parallel solutions CH-Constructor (Bühler and Nusser, 2024) and OSRM (Luxen and Vetter, 2024, 2011) based on Vetter’s algorithm. In tests on a 96-core machine with 16 graphs (Tab. 4), their performance is only up to 7.04×7.04\times7.04 × faster than a highly-optimized sequential implementation, and can be up to 3.59×\times× slower on some graphs.

In the next section, we present our solution that achieves a scalable and efficient parallel CH construction algorithm.

3. Our Parallel CH Construction Algorithm

Input: A graph G=(V,E)𝐺𝑉𝐸G=(V,E)italic_G = ( italic_V , italic_E ).
Output: A contraction hierarchy GCH=(V,ECH)subscript𝐺CH𝑉subscript𝐸CH\mathit{G_{\scriptsize\mbox{\it CH}}}=(V,\mathit{\mathit{E_{\scriptsize\mbox{% \it CH}}}})italic_G start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT = ( italic_V , italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT )
Maintains : The overlay graph GO=(VO,EO)subscript𝐺Osubscript𝑉Osubscript𝐸O{\mathit{G_{\scriptsize\mbox{\it O}}}}=(\mathit{V_{\scriptsize\mbox{\it O}}},{% \mathit{E_{\scriptsize\mbox{\it O}}}})italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT = ( italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT , italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT ). Initially GO=Gsubscript𝐺O𝐺{\mathit{G_{\scriptsize\mbox{\it O}}}}=Gitalic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT = italic_G.
The score of each vertex P⁢[⋅]𝑃delimited-[]⋅\mathit{P}[\cdot]italic_P [ ⋅ ].
1 ECH←∅←subscript𝐸CH\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}}\leftarrow\emptysetitalic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT ← ∅; VA←V←subscript𝑉𝐴𝑉\mathit{V_{A}}\leftarrow Vitalic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT ← italic_V; VW←V←subscript𝑉𝑊𝑉\mathit{V_{W}}\leftarrow Vitalic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT ← italic_V
2while VO≠∅subscript𝑉O\mathit{V_{\scriptsize\mbox{\it O}}}\neq\emptysetitalic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT ≠ ∅ do
       // Step 1: Run witness path searches (WPSes) from all vertices in VWsubscriptVW\mathit{V_{W}}italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT to prepare for score recomputation for all vertices in VAsubscriptVA\mathit{V_{A}}italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT
3       LocalSearch(VW,VA,GO)subscript𝑉𝑊subscript𝑉𝐴subscript𝐺O(\mathit{V_{W}},\mathit{V_{A}},{\mathit{G_{\scriptsize\mbox{\it O}}}})( italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT , italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT , italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT )
      // Step 2: For each vertex v∈VAvsubscriptVAv\in\mathit{V_{A}}italic_v ∈ italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT, (re)compute its score P⁢[v]Pdelimited-[]v\mathit{P}[v]italic_P [ italic_v ]
4       Score(VA,GO)subscript𝑉𝐴subscript𝐺O(\mathit{V_{A}},{\mathit{G_{\scriptsize\mbox{\it O}}}})( italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT , italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT )
      // Step 3: Select an independent set of vertices VF⊆VOsubscriptVFsubscriptVO\mathit{V_{F}}\subseteq\mathit{V_{\scriptsize\mbox{\it O}}}italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT ⊆ italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT to be contracted based on the score array P⁢[⋅]Pdelimited-[]⋅P[\cdot]italic_P [ ⋅ ]
5       VF←←subscript𝑉𝐹absent\mathit{V_{F}}\leftarrowitalic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT ←Select(GO,P)subscript𝐺O𝑃({\mathit{G_{\scriptsize\mbox{\it O}}}},P)( italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT , italic_P )
      // Step 4: Contract vertices in VFsubscriptVF\mathit{V_{F}}italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT; update ECHsubscriptECH\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}}italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT and GOsubscriptGO{\mathit{G_{\scriptsize\mbox{\it O}}}}italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT accordingly; return LSM as the sources of WPSes in the next iteration
6       VW←←subscript𝑉𝑊absent\mathit{V_{W}}\leftarrowitalic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT ←Contract(VF,GO)subscript𝑉𝐹subscript𝐺O(\mathit{V_{F}},{\mathit{G_{\scriptsize\mbox{\it O}}}})( italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT , italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT )
       // VAsubscript𝑉𝐴\mathit{V_{A}}italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT in the next round includes all neighbors of VFsubscript𝑉𝐹\mathit{V_{F}}italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT
7       VA←N⁢(VF)←subscript𝑉𝐴𝑁subscript𝑉𝐹\mathit{V_{A}}\leftarrow N(\mathit{V_{F}})italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT ← italic_N ( italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT )
return Postprocess(V,ECH)𝑉subscript𝐸CH(V,\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}})( italic_V , italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT )
 // reindexing
8
Algorithm 2 Our Parallel Algorithm for CH Construction

In this section, we present SPoCH for generating contraction hierarchies GCH=(V,ECH)subscript𝐺CH𝑉subscript𝐸CH\mathit{G_{\scriptsize\mbox{\it CH}}}=(V,\mathit{\mathit{E_{\scriptsize\mbox{% \it CH}}}})italic_G start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT = ( italic_V , italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT ) efficiently in parallel. As with other parallel CH algorithms, SPoCH also follows the high-level idea from Vetter (Vetter, 2009) that contracts a batch of vertices in an IS in parallel. To achieve high parallelism without sacrificing CH quality, SPoCH uses several novel techniques that will be discussed in this section.

We present the pseudocode of SPoCH in Alg. 2. It takes a graph G=(V,E)𝐺𝑉𝐸G=(V,E)italic_G = ( italic_V , italic_E ) as the input, and returns the contracted graph GCHsubscript𝐺CH\mathit{G_{\scriptsize\mbox{\it CH}}}italic_G start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT. Note that the vertex set in GCHsubscript𝐺CH\mathit{G_{\scriptsize\mbox{\it CH}}}italic_G start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT is the same as in the input graph, and the algorithm only needs to compute the edges ECHsubscript𝐸CH\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}}italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT of the contracted graph. Similar to Alg. 1, we also maintain the overlay graph GO=(VO,EO)subscript𝐺Osubscript𝑉Osubscript𝐸O{\mathit{G_{\scriptsize\mbox{\it O}}}}=(\mathit{V_{\scriptsize\mbox{\it O}}},{% \mathit{E_{\scriptsize\mbox{\it O}}}})italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT = ( italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT , italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT ) to reflect the changes in the original graph due to contractions.

The main loop of our algorithm (the while-loop on line 2) repeatedly finds an IS of vertices and contracts them. We summarize this process in five steps. We will briefly outline the high-level idea here, and then elaborate on each step.

One of the key insights in SPoCH is the introduction of the LocalSearch step on line 2. This step is used to reduce redundant WPSes, create more parallel tasks, and overall reduce computation while improving parallelism. With the details of this step described in Sec. 3.1, the overall goal of this step is to preprocess the current overlay graph by running WPSes from a set of vertices VWsubscript𝑉𝑊\mathit{V_{W}}italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT, in order to (re)compute the score for a set of vertices VAsubscript𝑉𝐴\mathit{V_{A}}italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT. At the beginning, both VWsubscript𝑉𝑊\mathit{V_{W}}italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT and VAsubscript𝑉𝐴\mathit{V_{A}}italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT are V𝑉Vitalic_V. In later rounds, VWsubscript𝑉𝑊\mathit{V_{W}}italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT will be computed by the Contract step in the previous round, and VAsubscript𝑉𝐴\mathit{V_{A}}italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT are the neighbors of vertices contracted in the previous round. In Vetter’s algorithm, WPSes are directly performed when recomputing scores in each round. SPoCH separates this part with new designs, and further utilizes the results for more optimizations.

The next two steps Score and Select will (re-)calculate the score for vertices in VAsubscript𝑉𝐴\mathit{V_{A}}italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT, and choose an independent set of vertices VFsubscript𝑉𝐹\mathit{V_{F}}italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT to contract. To ensure that selected vertices have low scores, VFsubscript𝑉𝐹\mathit{V_{F}}italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT includes all vertices with the lowest score in their neighborhood. Importantly, the Score step makes use of the results from the LocalSearch step to avoid redundant computation.

Finally, the Contract step performs the contraction and updates the overlay graph, which is another major improvement over existing solutions. In this step, all incident edges to VFsubscript𝑉𝐹\mathit{V_{F}}italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT are moved to GCHsubscript𝐺CH\mathit{G_{\scriptsize\mbox{\it CH}}}italic_G start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT, and the corresponding shortcuts are added to GOsubscript𝐺O{\mathit{G_{\scriptsize\mbox{\it O}}}}italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT. All these updates in SPoCH are highly parallel. Especially, we use lock-free data structures to update the graph efficiently. We introduce the algorithmic idea in Sec. 3.4, and discuss the parallel data structure support in Sec. 4. In addition, to facilitate the LocalSearch step, the Contract step will generate the set of vertices VWsubscript𝑉𝑊\mathit{V_{W}}italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT, which are all vertices from which we will start WPS in the next round.

VFsubscript𝑉𝐹\mathit{V_{F}}italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT: Vertices to be contracted VF={v:P⁢[v]<P⁢[u]⁢∀u∈N⁢(v)}subscript𝑉𝐹conditional-set𝑣𝑃delimited-[]𝑣𝑃delimited-[]𝑢for-all𝑢𝑁𝑣\mathit{V_{F}}=\{v:\mathit{P}[v]<\mathit{P}[u]~{}\forall u\in N(v)\}italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT = { italic_v : italic_P [ italic_v ] < italic_P [ italic_u ] ∀ italic_u ∈ italic_N ( italic_v ) }
VAsubscript𝑉𝐴\mathit{V_{A}}italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT: Vertices that need to VA=N⁢(VF)subscript𝑉𝐴𝑁subscript𝑉𝐹\mathit{V_{A}}=N(\mathit{V_{F}})italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT = italic_N ( italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT ) in the previous round
recalculate their scores
E+superscript𝐸\mathit{E^{+}}italic_E start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT: Shortcuts to be added
VSsubscript𝑉𝑆\mathit{V_{S}}italic_V start_POSTSUBSCRIPT italic_S end_POSTSUBSCRIPT: Starting points of shortcuts VS={v1:(v1,v2)∈E+}subscript𝑉𝑆conditional-setsubscript𝑣1subscript𝑣1subscript𝑣2superscript𝐸\mathit{V_{S}}=\{v_{1}:(v_{1},v_{2})\in\mathit{E^{+}}\}italic_V start_POSTSUBSCRIPT italic_S end_POSTSUBSCRIPT = { italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT : ( italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ) ∈ italic_E start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT }
VWsubscript𝑉𝑊\mathit{V_{W}}italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT: Set of WPS sources VW=VS∪Nin⁢(VS)subscript𝑉𝑊subscript𝑉𝑆subscript𝑁insubscript𝑉𝑆\mathit{V_{W}}=\mathit{V_{S}}\cup N_{\scriptsize\mbox{\it in}}(\mathit{V_{S}})italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT = italic_V start_POSTSUBSCRIPT italic_S end_POSTSUBSCRIPT ∪ italic_N start_POSTSUBSCRIPT in end_POSTSUBSCRIPT ( italic_V start_POSTSUBSCRIPT italic_S end_POSTSUBSCRIPT )
Table 3. Additional notations used in our algorithm.

We summarize the notations in our algorithm in Tab. 3. Next, we will elaborate on each step in Alg. 2. Later, in Sec. 4, we discuss using efficient data structures to support this algorithm.

3.1. Step 1: Local Search with Memoization

1 Function LocalSearch(VW,VA,GO=(VO,EO))subscript𝑉𝑊subscript𝑉𝐴subscript𝐺Osubscript𝑉Osubscript𝐸O(\mathit{V_{W}},\mathit{V_{A}},{\mathit{G_{\scriptsize\mbox{\it O}}}}=(\mathit% {V_{\scriptsize\mbox{\it O}}},{\mathit{E_{\scriptsize\mbox{\it O}}}}))( italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT , italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT , italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT = ( italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT , italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT ) )
2       ParallelForEach s𝑠sitalic_s in VWsubscript𝑉𝑊\mathit{V_{W}}italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT do
             Vn=∅subscript𝑉𝑛V_{n}=\emptysetitalic_V start_POSTSUBSCRIPT italic_n end_POSTSUBSCRIPT = ∅
             // the set of vertices that need to be settled by s𝑠sitalic_s
3             foreach u∈Nout⁢(s)∧u∈VA𝑢subscript𝑁out𝑠𝑢subscript𝑉𝐴u\in N_{\scriptsize\mbox{\it out}}(s)\land u\in\mathit{V_{A}}italic_u ∈ italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_s ) ∧ italic_u ∈ italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT do
4                   foreach v∈Nout⁢(u):(v≠s)∧((s,v)∉𝑑𝑖𝑠𝑡):𝑣subscript𝑁out𝑢𝑣𝑠𝑠𝑣𝑑𝑖𝑠𝑡v\in N_{\scriptsize\mbox{\it out}}(u):(v\neq s)\land((s,v)\not\in\mathit{dist})italic_v ∈ italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_u ) : ( italic_v ≠ italic_s ) ∧ ( ( italic_s , italic_v ) ∉ italic_dist ) do
5                         Vn.𝑖𝑛𝑠𝑒𝑟𝑡⁢(v)formulae-sequencesubscript𝑉𝑛𝑖𝑛𝑠𝑒𝑟𝑡𝑣V_{n}.\mathit{insert}(v)italic_V start_POSTSUBSCRIPT italic_n end_POSTSUBSCRIPT . italic_insert ( italic_v )
6            [WPS from s𝑠sbold_italic_s] Run Dijkstra’s algorithm starting from s𝑠sitalic_s. Terminate the algorithm either all v∈Vn𝑣subscript𝑉𝑛v\in V_{n}italic_v ∈ italic_V start_POSTSUBSCRIPT italic_n end_POSTSUBSCRIPT are settled, or θ𝜃\thetaitalic_θ nodes are settled. Add the shortest distance w𝑤witalic_w from s𝑠sitalic_s to all v∈Vn𝑣subscript𝑉𝑛v\in V_{n}italic_v ∈ italic_V start_POSTSUBSCRIPT italic_n end_POSTSUBSCRIPT into parallel hashmap 𝑑𝑖𝑠𝑡𝑑𝑖𝑠𝑡\mathit{dist}italic_dist, so that 𝑑𝑖𝑠𝑡⁢[(s,v)]=w𝑑𝑖𝑠𝑡delimited-[]𝑠𝑣𝑤\mathit{dist}[(s,v)]=witalic_dist [ ( italic_s , italic_v ) ] = italic_w
7             foreach v𝑣vitalic_v: (s,v)∈EO𝑠𝑣subscript𝐸O(s,v)\in{\mathit{E_{\scriptsize\mbox{\it O}}}}( italic_s , italic_v ) ∈ italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT do
8                   if w⁢(s,v)>𝑑𝑖𝑠𝑡⁢[(s,v)]𝑤𝑠𝑣𝑑𝑖𝑠𝑡delimited-[]𝑠𝑣w(s,v)>\mathit{dist}[(s,v)]italic_w ( italic_s , italic_v ) > italic_dist [ ( italic_s , italic_v ) ] then delete edge (s,v)𝑠𝑣(s,v)( italic_s , italic_v ) from EOsubscript𝐸O{\mathit{E_{\scriptsize\mbox{\it O}}}}italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT
9                  
10      
Algorithm 3 Parallel Local Search

One improvement of SPoCH lies in the stand-alone LocalSearch step with memoization at the beginning of each round, presented in Alg. 3. This step runs witness path searches (WPSes) to prepare the shortest distances needed during scoring (Sec. 3.2) and contracting (Sec. 3.4). As mentioned in Sec. 2.1, to compute the score of a vertex v∈VA𝑣subscript𝑉𝐴v\in\mathit{V_{A}}italic_v ∈ italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT, a simulated contraction on v𝑣vitalic_v is needed to determine the shortcuts that would be added if we contract v𝑣vitalic_v. For each pair of vertices (u1,u2)subscript𝑢1subscript𝑢2(u_{1},u_{2})( italic_u start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_u start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ) where u1∈Nin⁢(v)subscript𝑢1subscript𝑁in𝑣u_{1}\in N_{\scriptsize\mbox{\it in}}(v)italic_u start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT ∈ italic_N start_POSTSUBSCRIPT in end_POSTSUBSCRIPT ( italic_v ) and u2∈Nout⁢(v)subscript𝑢2subscript𝑁out𝑣u_{2}\in N_{\scriptsize\mbox{\it out}}(v)italic_u start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ∈ italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_v ), if v𝑣vitalic_v is on the shortest paths from u1subscript𝑢1u_{1}italic_u start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT to u2subscript𝑢2u_{2}italic_u start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT, then a shortcut is needed. We also use the WPS as described in Sec. 2.1, which runs Dijkstra from u1subscript𝑢1u_{1}italic_u start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT until u2subscript𝑢2u_{2}italic_u start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT is settled to find the shortest distance between them.

Running WPSes is crucial to determine the score of a vertex. However, it is also expensive—to compute the edge difference for a vertex v𝑣vitalic_v, WPSes are needed on all pairs in Nin⁢(v)×Nout⁢(v)subscript𝑁in𝑣subscript𝑁out𝑣N_{\scriptsize\mbox{\it in}}(v)\times N_{\scriptsize\mbox{\it out}}(v)italic_N start_POSTSUBSCRIPT in end_POSTSUBSCRIPT ( italic_v ) × italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_v ) (Cartesian product). Fully finishing WPSes for all pairs may be expensive, especially when there exists one pair of vertices that are extremely far away. Therefore, most existing systems (including ours) also limit the size of each WPS to be θ𝜃\thetaitalic_θ, i.e., each Dijkstra’s algorithm only searches for θ𝜃\thetaitalic_θ iterations. In this case, some suboptimal shortcuts may be added and make the CH (necessarily) larger. To improve the overall performance, we highlight three key techniques in our LocalSearch step: batching, memoization, and pruning.

Batching.  Our first technique is to collect all WPSes needed and process them in a batch. In SPoCH, the Contract step generates a vertex set VWsubscript𝑉𝑊\mathit{V_{W}}italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT, which serves as sources of WPSes in the next round. By maintaining them in a batch, we process all WPSes in a batch to avoid redundant computation and achieve better parallelism. For example, if two simulated contractions trigger WPSes from the same vertex, they will be performed only once in the batch. An illustration of all WPS sources and how batching saves the number of WPSes is given in Fig. 2. Executing all WPSes as one batch allows us to optimize them as a whole, and provides greater opportunity to exploit parallelism.

Refer to caption
Figure 2. An illustration of notations and benefit of batching in SPoCH. The two figures illustrate the vertex sets VA,VF,VSsubscript𝑉𝐴subscript𝑉𝐹subscript𝑉𝑆\mathit{V_{A}},\mathit{V_{F}},\mathit{V_{S}}italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT , italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT , italic_V start_POSTSUBSCRIPT italic_S end_POSTSUBSCRIPT, and VWsubscript𝑉𝑊\mathit{V_{W}}italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT. A node in both purple and green means it is in both VWsubscript𝑉𝑊\mathit{V_{W}}italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT and VSsubscript𝑉𝑆\mathit{V_{S}}italic_V start_POSTSUBSCRIPT italic_S end_POSTSUBSCRIPT. On the right, we show an illustration of using batching to save WPSes. To rescore the four vertices in VAsubscript𝑉𝐴\mathit{V_{A}}italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT, for each v∈VA𝑣subscript𝑉𝐴v\in\mathit{V_{A}}italic_v ∈ italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT, previous solutions will perform WPSes on all pairs Nin⁢(v)×Nout⁢(v)subscript𝑁in𝑣subscript𝑁out𝑣N_{\scriptsize\mbox{\it in}}(v)\times N_{\scriptsize\mbox{\it out}}(v)italic_N start_POSTSUBSCRIPT in end_POSTSUBSCRIPT ( italic_v ) × italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_v ). SPoCH identifies all possible WPS sources in the Contract step in the previous round and collect them in VWsubscript𝑉𝑊\mathit{V_{W}}italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT. Therefore, the WPSes from the same source will be conducted only once. In this example, SPoCH only need 5 WPSes instead of 9.

Memoization.  SPoCH uses memoization to avoid computing the shortest distance on the same vertex pair multiple times. In SPoCH, a map 𝑑𝑖𝑠𝑡𝑑𝑖𝑠𝑡\mathit{dist}italic_dist is used to memoize pairwise shortest distances computed in previous WPSes. If a WPS is required on the same pair again, we directly use the memoized results without running another WPS. This information is reused in pruning (described below) and later steps Score and Contract.

Pruning.  A byproduct of memoizing the distances in 𝑑𝑖𝑠𝑡𝑑𝑖𝑠𝑡\mathit{dist}italic_dist is to prune redundant edges in GOsubscript𝐺O{\mathit{G_{\scriptsize\mbox{\it O}}}}italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT. As the algorithm proceeds, more pairs of vertices (u,v)𝑢𝑣(u,v)( italic_u , italic_v ) obtain their true shortest distance d⁢(u,v)𝑑𝑢𝑣d(u,v)italic_d ( italic_u , italic_v ) by WPSes (using Dijkstra’s algorithm). If there is an edge between u𝑢uitalic_u and v𝑣vitalic_v in the current overlay graph (possibly a suboptimal shortcut added in previous rounds) with weight w>d⁢(u,v)𝑤𝑑𝑢𝑣w>d(u,v)italic_w > italic_d ( italic_u , italic_v ), the edge can be deleted since it is redundant. This optimization effectively reduces the size of the overlay graph, which potentially improves the performance for both construction and query.

The LocalSearch Step.  With the three techniques, we now describe the LocalSearch step. The step takes VWsubscript𝑉𝑊\mathit{V_{W}}italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT, VAsubscript𝑉𝐴\mathit{V_{A}}italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT and GOsubscript𝐺O{\mathit{G_{\scriptsize\mbox{\it O}}}}italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT as input, and performs WPSes from all vertices in VWsubscript𝑉𝑊\mathit{V_{W}}italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT to recompute the scores for all vertices in VAsubscript𝑉𝐴\mathit{V_{A}}italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT. For each WPS from s𝑠sitalic_s, we first identify the set of vertices Vnsubscript𝑉𝑛V_{n}italic_V start_POSTSUBSCRIPT italic_n end_POSTSUBSCRIPT that need to be settled by s𝑠sitalic_s. On line 3–3 in Alg. 3, for each u∈Nout⁢(s)𝑢subscript𝑁out𝑠u\in N_{\scriptsize\mbox{\it out}}(s)italic_u ∈ italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_s ), if u∈VA𝑢subscript𝑉𝐴u\in\mathit{V_{A}}italic_u ∈ italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT, this means that s𝑠sitalic_s is a source of WPS because s𝑠sitalic_s is an in-neighbor of an affected vertex u𝑢uitalic_u. To compute the new score for u𝑢uitalic_u, the shortest distances from s𝑠sitalic_s to each vertex v∈Nout⁢(u)𝑣subscript𝑁out𝑢v\in N_{\scriptsize\mbox{\it out}}(u)italic_v ∈ italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_u ) need to be computed. Therefore, we collect Nout⁢(u)subscript𝑁out𝑢N_{\scriptsize\mbox{\it out}}(u)italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_u ), the out-neighbors of u𝑢uitalic_u, to the set Vnsubscript𝑉𝑛V_{n}italic_V start_POSTSUBSCRIPT italic_n end_POSTSUBSCRIPT. Note that s𝑠sitalic_s can correspond to multiple affected vertices u𝑢uitalic_u. In this case, all of their out-neighbors will be gathered in Vnsubscript𝑉𝑛V_{n}italic_V start_POSTSUBSCRIPT italic_n end_POSTSUBSCRIPT, and only incur one search starting from s𝑠sitalic_s. Before adding a vertex v𝑣vitalic_v to Vnsubscript𝑉𝑛V_{n}italic_V start_POSTSUBSCRIPT italic_n end_POSTSUBSCRIPT, we also check if (s,v)𝑠𝑣(s,v)( italic_s , italic_v ) has already been memoized in 𝑑𝑖𝑠𝑡𝑑𝑖𝑠𝑡\mathit{dist}italic_dist. If so, v𝑣vitalic_v will be skipped.

With all target vertices of WPS originating from s𝑠sitalic_s gathered in Vnsubscript𝑉𝑛V_{n}italic_V start_POSTSUBSCRIPT italic_n end_POSTSUBSCRIPT, We run Dijkstra’s algorithm from s𝑠sitalic_s to search for all vertices in Vnsubscript𝑉𝑛V_{n}italic_V start_POSTSUBSCRIPT italic_n end_POSTSUBSCRIPT. During the algorithm, the shortest distances from s𝑠sitalic_s to each v∈Vn𝑣subscript𝑉𝑛v\in V_{n}italic_v ∈ italic_V start_POSTSUBSCRIPT italic_n end_POSTSUBSCRIPT are inserted into 𝑑𝑖𝑠𝑡𝑑𝑖𝑠𝑡\mathit{dist}italic_dist. As mentioned, a WPS terminates either when all vertices in Vnsubscript𝑉𝑛V_{n}italic_V start_POSTSUBSCRIPT italic_n end_POSTSUBSCRIPT have been settled, or when θ𝜃\thetaitalic_θ vertices have been settled (so that the overall cost is limited). If the WPS terminates before a vertex v∈Vn𝑣subscript𝑉𝑛v\in V_{n}italic_v ∈ italic_V start_POSTSUBSCRIPT italic_n end_POSTSUBSCRIPT is settled, the distance between s𝑠sitalic_s and v𝑣vitalic_v will be viewed as ∞\infty∞ and a shortcut between s𝑠sitalic_s and v𝑣vitalic_v is always generated.

3.2. Step 2: Score

1 Function Score(VA,GO=(VO,EO))subscript𝑉𝐴subscript𝐺Osubscript𝑉Osubscript𝐸O(\mathit{V_{A}},{\mathit{G_{\scriptsize\mbox{\it O}}}}=(\mathit{V_{\scriptsize% \mbox{\it O}}},{\mathit{E_{\scriptsize\mbox{\it O}}}}))( italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT , italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT = ( italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT , italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT ) )
2       ParallelForEach u𝑢uitalic_u in VAsubscript𝑉𝐴\mathit{V_{A}}italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT do
3             P⁢[u]←−|N⁢(u)|←𝑃delimited-[]𝑢𝑁𝑢\mathit{P}[u]\leftarrow-|N(u)|italic_P [ italic_u ] ← - | italic_N ( italic_u ) |
4            foreach v1∈Nin⁢(u)subscript𝑣1subscript𝑁in𝑢v_{1}\in N_{\scriptsize\mbox{\it in}}(u)italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT ∈ italic_N start_POSTSUBSCRIPT in end_POSTSUBSCRIPT ( italic_u ) do
5                   foreach v2∈Nout⁢(u)subscript𝑣2subscript𝑁out𝑢v_{2}\in N_{\scriptsize\mbox{\it out}}(u)italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ∈ italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_u ) do
6                         if w⁢(v1,u)+w⁢(u,v2)=𝑑𝑖𝑠𝑡⁢⟨v1,v2⟩∨(v1,v2)∉𝑑𝑖𝑠𝑡𝑤subscript𝑣1𝑢𝑤𝑢subscript𝑣2𝑑𝑖𝑠𝑡subscript𝑣1subscript𝑣2subscript𝑣1subscript𝑣2𝑑𝑖𝑠𝑡w(v_{1},u)+w(u,v_{2})=\mathit{dist}\langle v_{1},v_{2}\rangle\lor(v_{1},v_{2})% \not\in\mathit{dist}italic_w ( italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_u ) + italic_w ( italic_u , italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ) = italic_dist ⟨ italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ⟩ ∨ ( italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ) ∉ italic_dist then  P⁢[u]←P⁢[u]+1←𝑃delimited-[]𝑢𝑃delimited-[]𝑢1\mathit{P}[u]\leftarrow\mathit{P}[u]+1italic_P [ italic_u ] ← italic_P [ italic_u ] + 1
7                        
Algorithm 4 Parallel Node Scoring

The second step computes the scores for each vertex u∈VA𝑢subscript𝑉𝐴u\in\mathit{V_{A}}italic_u ∈ italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT, as presented in Alg. 4. In the previous step, we have memoized the distances in the neighborhood of each u∈VA𝑢subscript𝑉𝐴u\in\mathit{V_{A}}italic_u ∈ italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT in a map d⁢i⁢s⁢t𝑑𝑖𝑠𝑡distitalic_d italic_i italic_s italic_t. To recompute the score of u∈VA𝑢subscript𝑉𝐴u\in\mathit{V_{A}}italic_u ∈ italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT, we first set P⁢[u]𝑃delimited-[]𝑢\mathit{P}[u]italic_P [ italic_u ] to the negative of its number of direct neighbors, i.e., −|N⁢(u)|𝑁𝑢-|N(u)|- | italic_N ( italic_u ) |, since |N⁢(u)|𝑁𝑢|N(u)|| italic_N ( italic_u ) | edges can be removed from GOsubscript𝐺O{\mathit{G_{\scriptsize\mbox{\it O}}}}italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT if u𝑢uitalic_u gets contracted (line 4). Then, we iterate through each pair of in-neighbors and out-neighbors of u𝑢uitalic_u and check if their distance through u𝑢uitalic_u is equal to or smaller than the shortest distance stored in 𝑑𝑖𝑠𝑡𝑑𝑖𝑠𝑡\mathit{dist}italic_dist. If so, we add one to P⁢[u]𝑃delimited-[]𝑢\mathit{P}[u]italic_P [ italic_u ] since one shortcut needs to be added.

For simplicity, our pseudocode only computes the edge difference, which is the most important component of the score for each vertex. In practice, many other components are considered in existing work. SPoCH uses a similar score function as in previous work, which is a combination of edge difference, vertex degree, hop distance, and hierarchy depth.

3.3. Step 3: Select

1 Function Select(GO=(VO,EO))subscript𝐺Osubscript𝑉Osubscript𝐸O({\mathit{G_{\scriptsize\mbox{\it O}}}}=(\mathit{V_{\scriptsize\mbox{\it O}}},% {\mathit{E_{\scriptsize\mbox{\it O}}}}))( italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT = ( italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT , italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT ) )
2       VF←∅←subscript𝑉𝐹\mathit{V_{F}}\leftarrow\emptysetitalic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT ← ∅
3      
4      ParallelForEach u𝑢uitalic_u in VOsubscript𝑉O\mathit{V_{\scriptsize\mbox{\it O}}}italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT do
5             if (∀v∈N⁢(u),P⁢[u]<P⁢[v])formulae-sequencefor-all𝑣𝑁𝑢𝑃delimited-[]𝑢𝑃delimited-[]𝑣(\forall v\in N(u),\mathit{P}[u]<\mathit{P}[v])( ∀ italic_v ∈ italic_N ( italic_u ) , italic_P [ italic_u ] < italic_P [ italic_v ] ) then  VF.insert⁢(u)formulae-sequencesubscript𝑉𝐹insert𝑢\mathit{V_{F}}.\text{insert}(u)italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT . insert ( italic_u )
6            
7       return VFsubscript𝑉𝐹\mathit{V_{F}}italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT
Algorithm 5 Parallel Node Selection

The Select step identifies a subset of vertices VFsubscript𝑉𝐹\mathit{V_{F}}italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT to contract in parallel, as presented in Alg. 5. We aim to contract multiple vertices in an independent set (i.e., do not share edges) simultaneously while prioritizing those with lower scores. To do this, we select all vertices that have the minimum score in its neighborhood. As shown in line 5–5 in Alg. 5, we process each vertex u𝑢uitalic_u in VOsubscript𝑉O\mathit{V_{\scriptsize\mbox{\it O}}}italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT. If u𝑢uitalic_u has the lowest score among its neighbors, we insert u𝑢uitalic_u into a set VFsubscript𝑉𝐹\mathit{V_{F}}italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT. To handle equal scores, we use the label of each vertex to break ties. This approach can find an independent set so that vertices with lower scores are contracted before those with higher scores. Fig. 1(b) presents an example of contracting a graph in parallel. In each round, multiple independent vertices can be contracted together, and the levels can be determined accordingly. The final graph preserves the shortest distance between any two vertices.

Selecting an IS of vertices ensures that shortcuts are always established between non-contracted vertices. The process is also reasonably fast for contraction. In our experiments, we observe that usually within 10–20 rounds, more than 99% of vertices are contracted, indicating high potential of parallelism in this algorithm.

3.4. Step 4: Contract

1 Function Contract(VF,GO=(VO,EO))subscript𝑉𝐹subscript𝐺Osubscript𝑉Osubscript𝐸O(\mathit{V_{F}},{\mathit{G_{\scriptsize\mbox{\it O}}}}=(\mathit{V_{\scriptsize% \mbox{\it O}}},{\mathit{E_{\scriptsize\mbox{\it O}}}}))( italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT , italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT = ( italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT , italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT ) )
2       VS←∅←subscript𝑉𝑆\mathit{V_{S}}\leftarrow\emptysetitalic_V start_POSTSUBSCRIPT italic_S end_POSTSUBSCRIPT ← ∅
3       ParallelForEach u𝑢uitalic_u in VFsubscript𝑉𝐹\mathit{V_{F}}italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT do
4             Remove u𝑢uitalic_u from VOsubscript𝑉O\mathit{V_{\scriptsize\mbox{\it O}}}italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT
5             foreach v∈Nin⁢(u)𝑣subscript𝑁in𝑢v\in N_{\scriptsize\mbox{\it in}}(u)italic_v ∈ italic_N start_POSTSUBSCRIPT in end_POSTSUBSCRIPT ( italic_u ) do  ECH↓.𝑖𝑛𝑠𝑒𝑟𝑡⁢(⟨v,u,w⁢(v,u)⟩)formulae-sequencesuperscriptsubscript𝐸CH↓𝑖𝑛𝑠𝑒𝑟𝑡𝑣𝑢𝑤𝑣𝑢\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}^{\downarrow}}.\mathit{insert}(% \langle v,u,w(v,u)\rangle)italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT start_POSTSUPERSCRIPT ↓ end_POSTSUPERSCRIPT . italic_insert ( ⟨ italic_v , italic_u , italic_w ( italic_v , italic_u ) ⟩ )
6             foreach v∈Nout⁢(u)𝑣subscript𝑁out𝑢v\in N_{\scriptsize\mbox{\it out}}(u)italic_v ∈ italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_u ) do  ECH↑.𝑖𝑛𝑠𝑒𝑟𝑡⁢(⟨u,v,w⁢(u,v)⟩)formulae-sequencesuperscriptsubscript𝐸CH↑𝑖𝑛𝑠𝑒𝑟𝑡𝑢𝑣𝑤𝑢𝑣\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}^{\uparrow}}.\mathit{insert}(% \langle u,v,w(u,v)\rangle)italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT start_POSTSUPERSCRIPT ↑ end_POSTSUPERSCRIPT . italic_insert ( ⟨ italic_u , italic_v , italic_w ( italic_u , italic_v ) ⟩ )
7            
8            for v1∈Nin⁢(u)subscript𝑣1subscript𝑁in𝑢v_{1}\in N_{\scriptsize\mbox{\it in}}(u)italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT ∈ italic_N start_POSTSUBSCRIPT in end_POSTSUBSCRIPT ( italic_u ) do
9                   for v2∈Nout⁢(u)subscript𝑣2subscript𝑁out𝑢v_{2}\in N_{\scriptsize\mbox{\it out}}(u)italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ∈ italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_u ) do
10                         if w⁢(v1,u)+w⁢(u,v2)=𝑑𝑖𝑠𝑡⁢[(v1,v2)]∨(v1,v2)∉𝑑𝑖𝑠𝑡𝑤subscript𝑣1𝑢𝑤𝑢subscript𝑣2𝑑𝑖𝑠𝑡delimited-[]subscript𝑣1subscript𝑣2subscript𝑣1subscript𝑣2𝑑𝑖𝑠𝑡w(v_{1},u)+w(u,v_{2})=\mathit{dist}[(v_{1},v_{2})]\lor(v_{1},v_{2})\not\in% \mathit{dist}italic_w ( italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_u ) + italic_w ( italic_u , italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ) = italic_dist [ ( italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ) ] ∨ ( italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ) ∉ italic_dist then
11                               VS.𝑖𝑛𝑠𝑒𝑟𝑡⁢(v1)formulae-sequencesubscript𝑉𝑆𝑖𝑛𝑠𝑒𝑟𝑡subscript𝑣1\mathit{V_{S}}.\mathit{insert}(v_{1})italic_V start_POSTSUBSCRIPT italic_S end_POSTSUBSCRIPT . italic_insert ( italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT )
12                               E+.𝑖𝑛𝑠𝑒𝑟𝑡⁢(⟨v1,v2,w⁢(v1,u)+w⁢(u,v2)⟩)formulae-sequencesuperscript𝐸𝑖𝑛𝑠𝑒𝑟𝑡subscript𝑣1subscript𝑣2𝑤subscript𝑣1𝑢𝑤𝑢subscript𝑣2\mathit{E^{+}}.\mathit{insert}(\langle v_{1},v_{2},w(v_{1},u)+w(u,v_{2})\rangle)italic_E start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT . italic_insert ( ⟨ italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT , italic_w ( italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_u ) + italic_w ( italic_u , italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ) ⟩ )
13             Remove the incident edges of u𝑢uitalic_u from EOsubscript𝐸O{\mathit{E_{\scriptsize\mbox{\it O}}}}italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT
14      EO=EO⁢⋃E+subscript𝐸Osubscript𝐸Osuperscript𝐸{\mathit{E_{\scriptsize\mbox{\it O}}}}={\mathit{E_{\scriptsize\mbox{\it O}}}}% \bigcup\mathit{E^{+}}italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT = italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT ⋃ italic_E start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT
15      VW←∅←subscript𝑉𝑊\mathit{V_{W}}\leftarrow\emptysetitalic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT ← ∅
16       ParallelForEach u𝑢uitalic_u in VSsubscript𝑉𝑆\mathit{V_{S}}italic_V start_POSTSUBSCRIPT italic_S end_POSTSUBSCRIPT do
17             VW.insert⁢(u)formulae-sequencesubscript𝑉𝑊insert𝑢\mathit{V_{W}}.\text{insert}(u)italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT . insert ( italic_u )
18            ParallelForEach v∈Nin⁢(u)𝑣subscript𝑁in𝑢v\in N_{\scriptsize\mbox{\it in}}(u)italic_v ∈ italic_N start_POSTSUBSCRIPT in end_POSTSUBSCRIPT ( italic_u ) do  VW.insert⁢(v)formulae-sequencesubscript𝑉𝑊insert𝑣\mathit{V_{W}}.\text{insert}(v)italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT . insert ( italic_v )
19            
20      return VWsubscript𝑉𝑊\mathit{V_{W}}italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT
Algorithm 6 Parallel Node Contraction

With the independent set VFsubscript𝑉𝐹\mathit{V_{F}}italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT computed in the previous round, the last step Contract performs the actual contraction, as given in Alg. 6. This step moves all vertices in VFsubscript𝑉𝐹\mathit{V_{F}}italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT from VOsubscript𝑉O\mathit{V_{\scriptsize\mbox{\it O}}}italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT to the CH, adds incident edges to ECHsubscript𝐸CH\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}}italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT, computes and adds the relevant shortcuts to EOsubscript𝐸O{\mathit{E_{\scriptsize\mbox{\it O}}}}italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT, and finally generates the set VWsubscript𝑉𝑊\mathit{V_{W}}italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT as the sources for running WPSes in the next round, which will be used in the next LocalSearch step.

In Alg. 6, we process each vertex u∈VF𝑢subscript𝑉𝐹u\in\mathit{V_{F}}italic_u ∈ italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT in parallel. We first remove each vertex u∈VF𝑢subscript𝑉𝐹u\in\mathit{V_{F}}italic_u ∈ italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT from the overlay vertex set VOsubscript𝑉O\mathit{V_{\scriptsize\mbox{\it O}}}italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT in line 6. Next, as shown in line 6–6 in Alg. 6, we move the incident edges of u𝑢uitalic_u to ECHsubscript𝐸CH\mathit{E_{\scriptsize\mbox{\it CH}}}italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT, which contains two subsets: ECH↓superscriptsubscript𝐸CH↓\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}^{\downarrow}}italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT start_POSTSUPERSCRIPT ↓ end_POSTSUPERSCRIPT and ECH↑superscriptsubscript𝐸CH↑\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}^{\uparrow}}italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT start_POSTSUPERSCRIPT ↑ end_POSTSUPERSCRIPT. When moving a vertex u𝑢uitalic_u to the CH, its incoming edges are stored in ECH↓superscriptsubscript𝐸CH↓\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}^{\downarrow}}italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT start_POSTSUPERSCRIPT ↓ end_POSTSUPERSCRIPT, and outgoing edges are stored in ECH↑superscriptsubscript𝐸CH↑\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}^{\uparrow}}italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT start_POSTSUPERSCRIPT ↑ end_POSTSUPERSCRIPT. This separation is necessary because during the shortest path queries, the search from the target only proceeds “backward” by following incoming edges in ECH↓superscriptsubscript𝐸CH↓\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}^{\downarrow}}italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT start_POSTSUPERSCRIPT ↓ end_POSTSUPERSCRIPT, while the search from the source only moves “forward” by following outgoing edges in ECH↑superscriptsubscript𝐸CH↑\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}^{\uparrow}}italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT start_POSTSUPERSCRIPT ↑ end_POSTSUPERSCRIPT. Both searches move “upward” in the CH.

Finally, as shown in line 6–6 in Alg. 6, for each vertex u𝑢uitalic_u to be contracted, we enumerate each in-neighbor v1∈Nin⁢(u)subscript𝑣1subscript𝑁in𝑢v_{1}\in N_{\scriptsize\mbox{\it in}}(u)italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT ∈ italic_N start_POSTSUBSCRIPT in end_POSTSUBSCRIPT ( italic_u ) and out-neighbor v2∈Nout⁢(u)subscript𝑣2subscript𝑁out𝑢v_{2}\in N_{\scriptsize\mbox{\it out}}(u)italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ∈ italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_u ). We establish a shortcut between v1subscript𝑣1v_{1}italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT and v2subscript𝑣2v_{2}italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT if necessary. This is performed by comparing w⁢(v1,u)+w⁢(u,v2)𝑤subscript𝑣1𝑢𝑤𝑢subscript𝑣2w(v_{1},u)+w(u,v_{2})italic_w ( italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_u ) + italic_w ( italic_u , italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ) with the shortest distance stored in 𝑑𝑖𝑠𝑡𝑑𝑖𝑠𝑡\mathit{dist}italic_dist. Recall that 𝑑𝑖𝑠𝑡𝑑𝑖𝑠𝑡\mathit{dist}italic_dist is a map generated in the LocalSearch step, which buffers shortest distances for relevant vertex pairs. If w⁢(v1,u)+w⁢(u,v2)=𝑑𝑖𝑠𝑡⁢[(v1,v2)]𝑤subscript𝑣1𝑢𝑤𝑢subscript𝑣2𝑑𝑖𝑠𝑡delimited-[]subscript𝑣1subscript𝑣2w(v_{1},u)+w(u,v_{2})=\mathit{dist}[(v_{1},v_{2})]italic_w ( italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_u ) + italic_w ( italic_u , italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ) = italic_dist [ ( italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ) ], then the shortest path between v1subscript𝑣1v_{1}italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT and v2subscript𝑣2v_{2}italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT can be via u𝑢uitalic_u. A special case is when (v1,v2)subscript𝑣1subscript𝑣2(v_{1},v_{2})( italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ) is not in 𝑑𝑖𝑠𝑡𝑑𝑖𝑠𝑡\mathit{dist}italic_dist, which means that the shortest distance between them was not computed in the LocalSearch step. In both cases, before contracting u𝑢uitalic_u, a shortcut needs to be established between v1subscript𝑣1v_{1}italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT and v2subscript𝑣2v_{2}italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT with weight w⁢(v1,u)+w⁢(u,v2)𝑤subscript𝑣1𝑢𝑤𝑢subscript𝑣2w(v_{1},u)+w(u,v_{2})italic_w ( italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_u ) + italic_w ( italic_u , italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ) (line 6).

Refer to caption
Figure 3. An illustration of concurrent shortcut insertions in SPoCH. During the Contract step, for every feasible candidate u∈VF𝑢subscript𝑉𝐹u\in\mathit{V_{F}}italic_u ∈ italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT, a dedicated thread inserts shortcuts between each pair (v1,v2)subscript𝑣1subscript𝑣2(v_{1},v_{2})( italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ) with v1∈Nin⁢(u)subscript𝑣1subscript𝑁in𝑢v_{1}\in N_{\scriptsize\mbox{\it in}}(u)italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT ∈ italic_N start_POSTSUBSCRIPT in end_POSTSUBSCRIPT ( italic_u ) and v2∈Nout⁢(u)subscript𝑣2subscript𝑁out𝑢v_{2}\in N_{\scriptsize\mbox{\it out}}(u)italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ∈ italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_u ). In the example, three threads simultaneously attempt to insert edges (D,B)𝐷𝐵(D,B)( italic_D , italic_B ), (D,F)𝐷𝐹(D,F)( italic_D , italic_F ), (D,G)𝐷𝐺(D,G)( italic_D , italic_G ), which share the same source D𝐷Ditalic_D, potentially creating a race condition.

However, a write–write race condition may occur when multiple shortcuts that share the same source or target node are inserted in parallel. An example of this issue is shown in Fig. 3. Vertex D𝐷Ditalic_D has three out-neighbors that are being contracted, each introduces a new out-neighbor to D𝐷Ditalic_D. Since the contractions of these three original out-neighbors, {A,E,H}𝐴𝐸𝐻\{A,E,H\}{ italic_A , italic_E , italic_H }, run concurrently, the insertions of D𝐷Ditalic_D’s new out-neighbors, {B,F,G}𝐵𝐹𝐺\{B,F,G\}{ italic_B , italic_F , italic_G }, also happen in parallel. Therefore, we need a parallel data structure that supports frequent and concurrent insertions, while providing fast query performance. In our implementation, we use the parallel hash map E+superscript𝐸\mathit{E^{+}}italic_E start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT in Sec. 4.1 to buffer all shortcuts added in this round and later combine it with EOsubscript𝐸O{\mathit{E_{\scriptsize\mbox{\it O}}}}italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT on line 6. In Sec. 4.2 we will describe other optimizations to reduce the cost to combine E+superscript𝐸\mathit{E^{+}}italic_E start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT and EOsubscript𝐸O{\mathit{E_{\scriptsize\mbox{\it O}}}}italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT. Finally, we can also remove all edges incident on u∈VF𝑢subscript𝑉𝐹u\in\mathit{V_{F}}italic_u ∈ italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT from EOsubscript𝐸O{\mathit{E_{\scriptsize\mbox{\it O}}}}italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT after adding shortcuts. This is because these edges no longer contribute to the reduction of distances to any vertices that remain uncontracted.

During the contraction step, we also generate the set VWsubscript𝑉𝑊\mathit{V_{W}}italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT that contains all sources for the WPSes in the next step. When contracting a vertex u𝑢uitalic_u, we check all v1∈Nin⁢(u)subscript𝑣1subscript𝑁in𝑢v_{1}\in N_{\scriptsize\mbox{\it in}}(u)italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT ∈ italic_N start_POSTSUBSCRIPT in end_POSTSUBSCRIPT ( italic_u ) and v2∈Nout⁢(u)subscript𝑣2subscript𝑁out𝑢v_{2}\in N_{\scriptsize\mbox{\it out}}(u)italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ∈ italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_u ) to see whether a shortcut from v1subscript𝑣1v_{1}italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT to v2subscript𝑣2v_{2}italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT needs to be established. If so, the scores for both v1subscript𝑣1v_{1}italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT and v2subscript𝑣2v_{2}italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT are affected, since both have their neighborhoods changed. To reconsider the score of v2subscript𝑣2v_{2}italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT, WPSes are needed from v1subscript𝑣1v_{1}italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT (to find the distance from the new in-neighbor v1subscript𝑣1v_{1}italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT to all Nout⁢(v2)subscript𝑁outsubscript𝑣2N_{\scriptsize\mbox{\it out}}(v_{2})italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT )). To reconsider the score of v1subscript𝑣1v_{1}italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT, WPSes are needed from all in-neighbors of v1subscript𝑣1v_{1}italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT (to find their distances to the new out-neighbor v2subscript𝑣2v_{2}italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT). Hence, we first use a set VSsubscript𝑉𝑆\mathit{V_{S}}italic_V start_POSTSUBSCRIPT italic_S end_POSTSUBSCRIPT to buffer all vertices v1subscript𝑣1v_{1}italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT. At the end of this step, we construct the set VWsubscript𝑉𝑊\mathit{V_{W}}italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT, which includes all vertices in VSsubscript𝑉𝑆\mathit{V_{S}}italic_V start_POSTSUBSCRIPT italic_S end_POSTSUBSCRIPT and all in-neighbors of them. This set will be used as the sources for the WPSes in the LocalSearch step in the next round. An illustration of the vertex sets VSsubscript𝑉𝑆\mathit{V_{S}}italic_V start_POSTSUBSCRIPT italic_S end_POSTSUBSCRIPT and VWsubscript𝑉𝑊\mathit{V_{W}}italic_V start_POSTSUBSCRIPT italic_W end_POSTSUBSCRIPT is shown in Fig. 2.

3.5. Postprocessing

After contracting all vertices, we re-index and sort the vertices (Blelloch et al., 2010; Dong et al., 2024a) by their levels and scores to improve query performance. In the sequential CH algorithm (Geisberger et al., 2008), vertices are reordered by the order in which they are contracted. In PHAST (Group, 2024), vertices are grouped by levels to perform SSSP queries on the CH, where lower level vertices have a lower rank than those in higher levels. We combine the two re-indexing methods and use vertex scores to approximate the contraction order. For any two vertices, if they are at the same level, they are sorted by score from the lowest to the highest. Otherwise, they are sorted by level from the lowest to the highest.

This re-indexing ensures that the vertices are efficiently organized for query processing, where paths are traversed based on the contraction rank. Once the vertices are re-indexed, we generate the final CH from ECH↑superscriptsubscript𝐸CH↑\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}^{\uparrow}}italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT start_POSTSUPERSCRIPT ↑ end_POSTSUPERSCRIPT and ECH↓superscriptsubscript𝐸CH↓\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}^{\downarrow}}italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT start_POSTSUPERSCRIPT ↓ end_POSTSUPERSCRIPT.

4. Parallel Data Structures

In Sec. 3, we presented the algorithmic idea of SPoCH. We note that, however, to support high performance in practice, the implementation relies on efficient data structures to support the algorithm, especially in handling dynamic updates of the overlay graph. Note that the (overlay) graph is highly dynamic—as the contraction proceeds, the graph consistently undergoes vertex removals (removing contracted vertices), edge removals (removing edges incident on contracted vertices), and edge insertions (adding shortcuts). Each of these operations require high parallelism to enable good performance for the overall CH algorithm. This section provides an overview of the parallel data structures used in our implementation.

Especially, to facilitate graph traversal, the overlay graph is usually maintained by an array-based structure (e.g., the CSR introduced below), which maintains the neighbors of the same vertex in a consecutive chunk in the memory. When multiple vertices are contracted in parallel, shortcuts are also generated by multiple threads. Inserting them into the overlay graph in a concurrent manner is therefore challenging. In existing implementation, OSRM avoids such complications by only contracting vertices with the minimum score in its 2-hop neighborhood. In this case, for a vertex v𝑣vitalic_v in the overlay graph, only one contracted vertex may update the neighbor list of v𝑣vitalic_v, and this will be performed sequentially. However, by restricting all contracted vertices to be at least two hops away, OSRM can choose far fewer vertices in the independent set in each round. In our experiments, we observe that such restriction greatly increase the number of rounds and degraded the performance. In CH-Constructor, a lock-based data structure is used, which also limits parallelism. In this section, we introduce the parallel data structures used in SPoCH, such that they support updating the overlay graph in an efficient and flexible way.

4.1. Background of Existing Data Structures

SPoCH mainly uses two data structures to maintain graphs: Compressed Sparse Row (CSR) (Tinney and Walker, 1967) and phase-concurrent hash table (Shun and Blelloch, 2014). In this section, we introduce the background of these data structures and their interface. Then in Sec. 4.2, we describe how they are used to support our algorithm.

Compressed Sparse Row (CSR) (Tinney and Walker, 1967).  CSR is a widely used array-based adjacency list representation in graph algorithms (Dong et al., 2021, 2023; Liu et al., 2025; Wang et al., 2023; Dhulipala et al., 2024; Dong et al., 2024b; Dhulipala et al., 2021; Wheatman et al., 2024), consisting of an edge array of size |E|𝐸|E|| italic_E | and an offset array of size |V|𝑉|V|| italic_V |. The edge array is a concatenation of neighbor lists (along with the corresponding edge weights) for all vertices. The neighbor list for each vertex is always contiguous. The offset array stores the start position of each vertex’s neighborhood in the edge array. CSR provides fast lookups with good cache locality due to its contiguous memory layout. However, it does not support insertion or deletion without rebuilding the entire edge and offset arrays. Our input graph is given in the CSR format.

Phase-Concurrent Hash Table (Shun and Blelloch, 2014).  The phase-concurrent hash table is an array-based structure that supports efficient operations for an unordered set, including insertions, deletions, and lookups. We use it to support multiple variables in SPoCH.

Unlike traditional concurrent hash tables that allow arbitrary operations of any type to run concurrently, the phase-concurrent hash table only permits operations of the same type to proceed simultaneously. This design ensures both serialization and determinism (w.r.t. execution order). To insert a key-value pair (k,v)𝑘𝑣(k,v)( italic_k , italic_v ) into the hash table H𝐻Hitalic_H, the key k𝑘kitalic_k is first hashed to a random index in H𝐻Hitalic_H. If this index is occupied, linear probing is performed to find the next available slot. Once an empty index is found, it swaps in the new record using an atomic operation compare_and_swap. When multiple threads attempt to modify the same memory location, the atomic operation guarantees that only one thread succeeds. All other threads will fail in the operation and continue linear probing. A special case for insertion is to update the value for the k𝑘kitalic_k if it already exists in the hash table. In our implementation where the value can refer to the weight of a given edge, we update the value if the newly inserted value is smaller than the current value stored in the hash table. In particular, when inserting (k,v)𝑘𝑣(k,v)( italic_k , italic_v ), if we find that k𝑘kitalic_k is already in the table during linear probing, we use an atomic operation write_min to update the value and keep the lower one. The atomicity of this operation guarantees that only the minimum value among all concurrent modifications is retained.

Similarly, to look up a record with key k𝑘kitalic_k, the hash table hashes k𝑘kitalic_k to an initial index i𝑖iitalic_i. Then, three cases are considered: 1) If H⁢[i]𝐻delimited-[]𝑖H[i]italic_H [ italic_i ] is empty, then k𝑘kitalic_k does not exist in the hash table. 2) If the key at H⁢[i]𝐻delimited-[]𝑖H[i]italic_H [ italic_i ] is k𝑘kitalic_k, then the record with key k𝑘kitalic_k has been found. 3) If the key at H⁢[i]𝐻delimited-[]𝑖H[i]italic_H [ italic_i ] is not k𝑘kitalic_k, the algorithm uses linear probing to continue searching for the key in subsequent indices, until k𝑘kitalic_k is found or an empty slot is encountered. To delete a record with key k𝑘kitalic_k, the algorithm first locates k𝑘kitalic_k in H𝐻Hitalic_H. If k𝑘kitalic_k is present, the corresponding index is marked as a tombstone, which is not treated as empty but will not match any existing keys. If k𝑘kitalic_k is not found in H𝐻Hitalic_H, no further action is needed.

For simplicity, we will refer to the phase-concurrent hash table as the hash table in the following part of this section.

4.2. Using Parallel Data Structures in SPoCH

Maintaining the Overlay Graph Dynamically.  To facilitate graph traversal, the overlay graph GO=(VO,EO)subscript𝐺Osubscript𝑉Osubscript𝐸O{\mathit{G_{\scriptsize\mbox{\it O}}}}=(\mathit{V_{\scriptsize\mbox{\it O}}},{% \mathit{E_{\scriptsize\mbox{\it O}}}})italic_G start_POSTSUBSCRIPT O end_POSTSUBSCRIPT = ( italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT , italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT ) is maintained using a CSR, such that getting the neighbor list for each vertex is always efficient. The challenge is then to efficiently reflect edge insertions in the static CSR structure.

To efficiently support these updates in the overlay graph, we use an auxiliary edge set E+superscript𝐸\mathit{E^{+}}italic_E start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT to tentatively maintain the newly added edges. Recall that when a node u𝑢uitalic_u is contracted, pairwise edges from v1∈Nin⁢(u)subscript𝑣1subscript𝑁in𝑢v_{1}\in N_{\scriptsize\mbox{\it in}}(u)italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT ∈ italic_N start_POSTSUBSCRIPT in end_POSTSUBSCRIPT ( italic_u ) to v2∈Nout⁢(u)subscript𝑣2subscript𝑁out𝑢v_{2}\in N_{\scriptsize\mbox{\it out}}(u)italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ∈ italic_N start_POSTSUBSCRIPT out end_POSTSUBSCRIPT ( italic_u ) need to be added to the overlay graph if the shortest path from v1subscript𝑣1v_{1}italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT to v2subscript𝑣2v_{2}italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT passes through u𝑢uitalic_u. Instead of adding them directly to the CSR of the overlay graph, we first use E+superscript𝐸\mathit{E^{+}}italic_E start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT to buffer them. Once a shortcut is generated (line 6 in Alg. 6), we directly add the new edge to E+superscript𝐸\mathit{E^{+}}italic_E start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT. However, adding these edges to E+superscript𝐸\mathit{E^{+}}italic_E start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT in parallel is challenging due to contention—although contracting two adjacent vertices is avoided by selecting an IS, it is possible for multiple vertices to be contracted share the same neighbor. In this case, new edges will be concurrently added to this neighbor. Moreover, edges with the same endpoints (u,v)𝑢𝑣(u,v)( italic_u , italic_v ) could be added to the overlay graph by two different contracted vertices c1subscript𝑐1c_{1}italic_c start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT and c2subscript𝑐2c_{2}italic_c start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT (i.e., via u𝑢uitalic_u–c1subscript𝑐1c_{1}italic_c start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT–v𝑣vitalic_v and u𝑢uitalic_u–c2subscript𝑐2c_{2}italic_c start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT–v𝑣vitalic_v). Among all the same edges, only the one with the smallest weight should be kept.

To support these updates in the overlay graph, we use a parallel hash table to maintain EOsubscript𝐸O{\mathit{E_{\scriptsize\mbox{\it O}}}}italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT. Once a shortcut (v1,v2)subscript𝑣1subscript𝑣2(v_{1},v_{2})( italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ) with weight w𝑤witalic_w is generated, we insert the triple (v1,v2,w)subscript𝑣1subscript𝑣2𝑤(v_{1},v_{2},w)( italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT , italic_w ) to the hash table. To insert an edge (v1,v2,w)subscript𝑣1subscript𝑣2𝑤(v_{1},v_{2},w)( italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT , italic_w ), we first check whether the edge (v1,v2)subscript𝑣1subscript𝑣2(v_{1},v_{2})( italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT ) exists in the hash table. If so, we use write_min to update the weight atomically if w𝑤witalic_w is smaller than the stored weight. Otherwise, we insert (v1,v2,w)subscript𝑣1subscript𝑣2𝑤(v_{1},v_{2},w)( italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT , italic_w ) to the hash table. Here we use v1subscript𝑣1v_{1}italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT as the key, and we will explain the design choice later in this section.

Note that even if we first collect shortcuts in E+superscript𝐸\mathit{E^{+}}italic_E start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT and combine them with EOsubscript𝐸O{\mathit{E_{\scriptsize\mbox{\it O}}}}italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT to avoid frequently updating the CSR, it can still be expensive to do so every round. Especially in later rounds, only a few vertices are contracted, generating a small number of shortcuts. In this case, updating the CSR to add a few edges is inefficient. To handle this issue, we combine the two sets in a lazy manner. In particular, we do not clear and combine E+superscript𝐸\mathit{E^{+}}italic_E start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT with EOsubscript𝐸O{\mathit{E_{\scriptsize\mbox{\it O}}}}italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT until E+superscript𝐸\mathit{E^{+}}italic_E start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT gets overloaded (i.e., a constant fraction of slots are taken). We set the hash table size to the number of edges in the initial CSR. This guarantees that when updating the CSR, there must be a comparable number of edge insertions, and thus the cost can be amortized.

However, the belated combination of E+superscript𝐸\mathit{E^{+}}italic_E start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT and EOsubscript𝐸O{\mathit{E_{\scriptsize\mbox{\it O}}}}italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT also requires all edges in E+superscript𝐸\mathit{E^{+}}italic_E start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT to be visible to later graph traversal operations. To do this, when inserting an edge (v1,v2,w)subscript𝑣1subscript𝑣2𝑤(v_{1},v_{2},w)( italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT , italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT , italic_w ), we use v1subscript𝑣1v_{1}italic_v start_POSTSUBSCRIPT 1 end_POSTSUBSCRIPT as the key and (v2,w)subscript𝑣2𝑤(v_{2},w)( italic_v start_POSTSUBSCRIPT 2 end_POSTSUBSCRIPT , italic_w ) as the value. Using only the first endpoint as the key allows us to look up the neighbors of a given vertex by linear probing, and also has better cache locality. When traversing all neighbors of u𝑢uitalic_u in the overlay graph, we first process all neighbors of u𝑢uitalic_u in the CSR EOsubscript𝐸O{\mathit{E_{\scriptsize\mbox{\it O}}}}italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT. We then further look up u𝑢uitalic_u in the hash table for E+superscript𝐸\mathit{E^{+}}italic_E start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT, and use linear probing to get all edges incident on u𝑢uitalic_u. These provide the additional neighbors of u𝑢uitalic_u w.r.t. shortcuts associated with u𝑢uitalic_u that have not yet been incorporated into EOsubscript𝐸O{\mathit{E_{\scriptsize\mbox{\it O}}}}italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT.

To identify when the hash table is overloaded, at the end of each round, we use random sampling to estimate the size of the hash table. If the hash table is filled up by a constant fraction, a merge of E+superscript𝐸\mathit{E^{+}}italic_E start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT with EOsubscript𝐸O{\mathit{E_{\scriptsize\mbox{\it O}}}}italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT is triggered. During this process, we also remove all edges incident to the vertices that have been contracted. In other words, the edge removals in line 6 of Alg. 6 are also performed lazily until the next update on EOsubscript𝐸O{\mathit{E_{\scriptsize\mbox{\it O}}}}italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT. Finally, we clear the hash table.

The final edges in the CH, ECH↑superscriptsubscript𝐸CH↑\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}^{\uparrow}}italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT start_POSTSUPERSCRIPT ↑ end_POSTSUPERSCRIPT and ECH↓superscriptsubscript𝐸CH↓\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}^{\downarrow}}italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT start_POSTSUPERSCRIPT ↓ end_POSTSUPERSCRIPT are also maintained by the phase-concurrent hash tables. For both of them, only the insertion operation is needed, which can be performed concurrently using atomic operations.

Maintaining the Distance Mapping 𝒅𝒊𝒔𝒕𝒅𝒊𝒔𝒕\mathit{dist}bold_italic_dist.  Another important data structure used in our algorithm is 𝑑𝑖𝑠𝑡𝑑𝑖𝑠𝑡\mathit{dist}italic_dist, which memoizes shortest distances for certain pairs of vertices discovered in WPSes. In our implementation, we also use a phase-concurrent hash table. When a shortest distance between u𝑢uitalic_u and v𝑣vitalic_v is computed as w𝑤witalic_w, we add (u,v,w)𝑢𝑣𝑤(u,v,w)( italic_u , italic_v , italic_w ) to the hash table, where (u,v)𝑢𝑣(u,v)( italic_u , italic_v ) is the key and w𝑤witalic_w is the value. By setting up the mapping in the LocalSearch step, later Score and Contract steps can directly look up the distance instead of performing another simulated contraction.

Similar to the shortcut set E+superscript𝐸\mathit{E^{+}}italic_E start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT, the number of elements in the 𝑑𝑖𝑠𝑡𝑑𝑖𝑠𝑡\mathit{dist}italic_dist map can also change dramatically in different rounds. When most slots in the hash table are empty, clearing the map becomes expensive. To address this issue, similar to how we maintain E+superscript𝐸\mathit{E^{+}}italic_E start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT, we clear the hash table for 𝑑𝑖𝑠𝑡𝑑𝑖𝑠𝑡\mathit{dist}italic_dist lazily. At the end of the round, we use sampling to estimate the load factor of the hash table. Only when a constant fraction of the hash table is full do we clear the hash table.

5. Query on Contraction Hierarchies

In this section, we describe the standard query algorithm on the CH. After constructing the CH, we re-index and sort all vertices by their level and score. In Geisberger’s sequential CH algorithm (Geisberger et al., 2008), vertices are reordered based on the order to contract them. π⁢(v)𝜋𝑣\mathit{\pi}(v)italic_π ( italic_v ) is the order to contract v𝑣vitalic_v. In PHAST (Delling et al., 2013b), vertices are re-grouped by levels to perform SSSP queries on CH, where vertices in lower levels always have a lower rank than those in higher levels. In the query phase, all the edges in ECHsubscript𝐸CH\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}}italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT are divided into two parts and forming two graphs: the out-edges from u𝑢uitalic_u to higher-ranked vertices are referred to as the upward edges: ECH↑←{(u,v)|π⁢[u]<π⁢[v]}←superscriptsubscript𝐸CH↑conditional-set𝑢𝑣𝜋delimited-[]𝑢𝜋delimited-[]𝑣\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}^{\uparrow}}\leftarrow\{(u,v)~{}|% ~{}\mathit{\pi}[u]<\mathit{\pi}[v]\}italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT start_POSTSUPERSCRIPT ↑ end_POSTSUPERSCRIPT ← { ( italic_u , italic_v ) | italic_π [ italic_u ] < italic_π [ italic_v ] }, while the in-edges from higher-ranked vertices to u𝑢uitalic_u as the downward edges: ECH↓←{(v,u)|π⁢[u]<π⁢[v]}←superscriptsubscript𝐸CH↓conditional-set𝑣𝑢𝜋delimited-[]𝑢𝜋delimited-[]𝑣\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}^{\downarrow}}\leftarrow\{(v,u)~{% }|~{}\mathit{\pi}[u]<\mathit{\pi}[v]\}italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT start_POSTSUPERSCRIPT ↓ end_POSTSUPERSCRIPT ← { ( italic_v , italic_u ) | italic_π [ italic_u ] < italic_π [ italic_v ] }. For each edge connecting two vertices, only the endpoint of the lower rank stores this edge in GCHsubscript𝐺CH\mathit{G_{\scriptsize\mbox{\it CH}}}italic_G start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT.

CH boost up the efficiency of point to point query (s𝑠sitalic_s-t𝑡titalic_t query) significantly. During a point-to-point query from s𝑠sitalic_s to t𝑡titalic_t, a forward search from s𝑠sitalic_s is performed on ECH↑superscriptsubscript𝐸CH↑\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}^{\uparrow}}italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT start_POSTSUPERSCRIPT ↑ end_POSTSUPERSCRIPT and a backward search from t𝑡titalic_t is performed on ECH↓superscriptsubscript𝐸CH↓\mathit{\mathit{E_{\scriptsize\mbox{\it CH}}}^{\downarrow}}italic_E start_POSTSUBSCRIPT CH end_POSTSUBSCRIPT start_POSTSUPERSCRIPT ↓ end_POSTSUPERSCRIPT, meaning that only edges leading to higher rank vertices are considered. For each vertex v𝑣vitalic_v, the distance from s𝑠sitalic_s to v𝑣vitalic_v and the distance from v𝑣vitalic_v to t𝑡titalic_t is maintained as ds⁢(v)subscript𝑑𝑠𝑣d_{s}(v)italic_d start_POSTSUBSCRIPT italic_s end_POSTSUBSCRIPT ( italic_v ) and dt⁢(v)subscript𝑑𝑡𝑣d_{t}(v)italic_d start_POSTSUBSCRIPT italic_t end_POSTSUBSCRIPT ( italic_v ) and the estimated distance from s𝑠sitalic_s to t𝑡titalic_t through v𝑣vitalic_v is 𝑑𝑖𝑠𝑡s⁢(v)+𝑑𝑖𝑠𝑡t⁢(v)subscript𝑑𝑖𝑠𝑡𝑠𝑣subscript𝑑𝑖𝑠𝑡𝑡𝑣\mathit{dist}_{s}(v)+\mathit{dist}_{t}(v)italic_dist start_POSTSUBSCRIPT italic_s end_POSTSUBSCRIPT ( italic_v ) + italic_dist start_POSTSUBSCRIPT italic_t end_POSTSUBSCRIPT ( italic_v ). Once the tentative shortest distance μ𝜇\muitalic_μ from s𝑠sitalic_s to v𝑣vitalic_v is not larger than the minimum value of the priority queue, the search result is settled to μ𝜇\muitalic_μ.

6. Experiments

Graph Road Synthetic k𝑘kitalic_k-NN
CA AO AF NA AS EU TRCE BUB CHN7 CHN8 HT5 CH2 CH5 GL2 GL5 GL10
# vertices 3.34M 6.22M 33.5M 87.0M 95.7M 131M 16.0M 21.2M 10.0M 100M 929K 4.21M 4.21M 24.9M 24.9M 24.9M
# edges 4.20M 8.24M 44.8M 113M 123M 169M 48.0M 63.6M 10.0M 100M 4.64M 8.42M 21.0M 49.8M 124M 249M
Build (s) Ours* 1.00 2.25 7.32 23.1 23.5 32.9 25.2 34.7 1.77 14.9 0.68 0.87 115 2.95 9.78 41.6
OSRM* 4.60 26.8 60.0 307 286 521 218 292 6.77 71.2 2.95 4.38 28028 30.7 103 391
CC* 19.0 32.9 334 1527 1548 2121 1218 1680 19.0 218 11.4 11.0 4737 83.3 453 2627
RK 22.1 94.7 466 2466 2619 3783 1765 2509 38.7 848 11.2 15.5 14048 108 716 4175
PHAST 15.5 66.2 287 1341 1421 1865 1470 2053 23.5 285 8.79 9.20 11361 46.0 373 2388
Ours-seq 29.3 89.6 383 1411 1347 1823 1816 2620 65.4 751 18.9 26.1 7454 125 545 2676
OSRM-seq 33.1 183 503 3246 2839 4101 3817 5273 34.4 390 28.4 18.3 ¿30000 82.4 1062 12082
CC-seq 29.2 119 553 2907 2604 3574 3658 5002 35.1 423 23.7 20.5 17268 133 812 6855
Query (μ𝜇\muitalic_μs) *Ours 7.99 27.9 17.2 93.3 60.4 137 376 452 5.59 7.74 6.16 3.55 244 1.93 10.2 33.9
OSRM* 13.6 43.8 28.7 163 97.6 223 454 512 7.94 9.74 6.37 3.84 325 2.20 12.5 39.0
CC* 22.1 96.4 48.1 317 190 424 910 1010 7.60 9.07 6.73 3.33 783 2.11 17.2 85.2
RK 6.65 24.0 13.6 79.1 47.9 102 292 337 4.93 6.33 4.63 2.69 176 1.52 6.25 19.0
PHAST 12.6 39.2 26.7 138 84.4 169 430 482 7.73 9.23 6.98 4.09 278 2.37 11.8 33.5
Table 4. Comparison of build time (in seconds) and query time (in microseconds) for all tested implementations across all graphs. “Ours” === our algorithm. “OSRM” === Open Source Routing Machine (Luxen and Vetter, 2011). “CC” === CH-Constructor (Bühler and Nusser, 2024). “RK” === RoutingKit (Dibbelt et al., 2016). “PHAST” === PHAST (Delling et al., 2013b). Implementations with asterisks (*) are parallel, and their sequential running times (with suffices “-seq”) are also included in the table. We report the parallel running time of CC using eight threads, as this configuration yields the best performance due to its limited scalability. For each graph, the best build time and best query time are bolded and underlined.

6.1. Experimental Setup

Environment.  We run our experiments on a machine machine with four Intel Xeon Gold 6252 CPUs (96 cores and 192 hyperthreads) and 1.5TB of main memory. We use numactl -i all for parallel tests to interleave the memory pages across CPUs in a round-robin fashion.

We implement our algorithms in C++ using ParlayLib (Blelloch et al., 2020a) for fork-join parallelism and parallel primitives, which is recently used in many papers on parallel algorithms (Gu et al., 2023; Shen et al., 2022; Ben-David et al., 2016; Blelloch and Dobson, 2022). ParlayLib is an algorithmic library with parallel building blocks, and also include a scheduler for support fork-join parallelism. SPoCH uses some of them such as sorting from ParlayLib, and uses the ParlayLib scheduler by default, which gives good performance in general. ParlayLib infrastructure also supports alternative schedulers (e.g., TBB (Intel Threading Building Blocks, [n. d.]) and OpenCilk (OpenCilk, 2020)). A comparison of different schedulers for SPoCH is presented in Tab. 5.

Benchmark Dataset.  To evaluate the performance of CH construction algorithms, we use 16 directed graphs. This dataset includes six road graphs, four synthetic graphs and six k𝑘kitalic_k-NN graphs.

  • •

    Road graphs are the primary use application domain for CH algorithms. Here we use Central America (CA), Australia Oceania (AO), Europe (EU), North America (NA), Asia (AS), and Africa (AF) from OpenStreetMap (OpenStreetMap contributors, 2010). The edge weights are natural weights from the source data, which are up to 225superscript2252^{25}2 start_POSTSUPERSCRIPT 25 end_POSTSUPERSCRIPT.

  • •

    Synthetic graphs with relatively low average degrees and high diameters are selected because these characteristics are ideal for testing the performance of CH algorithms. Here we use hugebubbles-0020 (BUB) and hugetrace-0020 (TRCE) from (Sanders et al., 2014), which feature 2D adaptively refined triangular meshes. We assign random weights for BUB and TRCE from 0 to 105superscript10510^{5}10 start_POSTSUPERSCRIPT 5 end_POSTSUPERSCRIPT. Chain graphs have 107superscript10710^{7}10 start_POSTSUPERSCRIPT 7 end_POSTSUPERSCRIPT (CHN7) and 108superscript10810^{8}10 start_POSTSUPERSCRIPT 8 end_POSTSUPERSCRIPT (CHN8) vertices with random weights from 1 to 32. Whenever a vertex is contracted in the chain, exactly one shortcut is added. Therefore, chain graphs are primarily used to test the cost related to non-scoring steps.

  • •

    𝒌𝒌\boldsymbol{k}bold_italic_k-NN graphs are used to evaluate the performance among different average degrees. In k𝑘kitalic_k-NN graphs, each vertex has k𝑘kitalic_k out-going edges pointing to its k𝑘kitalic_k-nearest neighbors, excluding itself. As k𝑘kitalic_k increases, the complexity of the scoring process in CH construction also increases. We use Humidity and Temperature with k=5𝑘5k=5italic_k = 5 (HT5) (Huerta et al., 2016), Chemical with k=2,5𝑘25k=2,5italic_k = 2 , 5 (CH2, CH5) (Fonollosa et al., 2015; Wang et al., 2021), and GeoLife with k=2,5,10𝑘2510k=2,5,10italic_k = 2 , 5 , 10 (GL2, GL5, GL10) (Wang et al., 2021; Zheng et al., 2008). Edge weights are randomly assigned from 1 to 32.

When we compare the average performance across all graphs, we always report the geometric mean values across all graphs.

Refer to caption
Figure 4. Build time and query time of all tested baselines, normalized to our algorithm. Lower is better. Since all running times are normalized to our algorithm, “Ours” is always equal to one, represented by the horizontal red dotted line.

Baseline Competitors.  We call all existing algorithms that we compare to as baselines. We compare SPoCH to two of the most well-received sequential implementations (RoutingKit and PHAST) and two state-of-the-art parallel open-source implementations (CH-Constructor and OSRM), described as follows.

  • •

    OSRM (Luxen and Vetter, 2024, 2011): A parallel routing engine designed to calculate the fastest routes between locations in the OpenStreetMap (OpenStreetMap contributors, 2010).

  • •

    CH-Constructor (CC) (Bühler and Nusser, 2024): A parallel CH construction implementation. As shown in Fig. 7, CC always achieves the best performance with around 8 threads. Hence, the reported construction time for CC is on 8888 threads instead of 192192192192 threads.

  • •

    RoutingKit (RK) (contributors, 2024; Dibbelt et al., 2016): An efficient sequential C++ library that applies the CH for fast route planning.

  • •

    PHAST (Group, 2024; Delling et al., 2013b): A widely used algorithm for parallel single-source shortest paths (SSSP) using CH. Its implementation includes a sequential process for constructing the CH, and a parallel process for running SSSP. In our experiment, we only compare to their construction time, which is sequential.

The basic frameworks of both parallel baselines (CC and OSRM) are based on Vetter (Vetter, 2009), which supports node scoring and contracting in parallel. Each time they score a vertex, both algorithms always perform witness searches on all of its neighbors, and they do not record any distance information from these searches. This means that if scoring two vertices involves performing WPSes on the same vertex, both implementations will perform this search twice, greatly increasing the total computation.

Since our main focus is on the performance of constructing CH, we obtain the CH generated from all baselines, and always use the same query algorithm on these CH structures to enable a fair comparison of the CH quality. We choose to use the query algorithm in RK since we observe that it has the best performance in general.

6.2. Overall Performance

Setup.  We conducted experiments to compare performance in terms of construction time and query time with established baselines. Lower time consumption in both indicates better algorithm quality. Tab. 4 presents a comparison among SPoCH and all baselines. In Fig. 4, we normalize the performance of all baselines to that of SPoCH (with SPoCH’s numbers always set to 1). For construction, we also include the sequential running times of all parallel implementations in Tab. 4 to reflect the total work. For queries, for each graph, we randomly select 1000100010001000 pairs of vertices, run these queries sequentially, and report the average time and the average number of vertices processed in the queries. As mentioned, we apply the same query algorithm to all baselines.

Note that different construction algorithms may generate different CHs on the same input graph. In particular, a sequential algorithm always chooses the best node to contract in each iteration, while a parallel algorithm may choose to contract multiple vertices together. Conceptually, this means that many of the vertices are contracted earlier than they should be. Therefore, the quality of the CH may be sacrificed due to parallel construction. To measure the quality of the CHs generated by different algorithms, we use the query time as an indicator. More precisely, since all algorithms are using the same query algorithm, a lower query time indicates that the CH structure itself has higher quality.

Construction Time Evaluation. Fig. 4 and Tab. 4 show that our approach achieves significantly better construction performance than all baselines, while maintaining high CH graph quality. Even our sequential construction time is competitive to the highly-optimized sequential baselines, which is within 1.7×\times× slower than RK and 2.9×\times× than PHAST, and can also be faster than them by up to 2.1×\times×. This indicates that our algorithm incurs very small overhead in work to enable parallelism. When running in parallel, the construction time of our algorithm is significantly faster than all baselines. Compared to the best parallel baseline OSRM, SPoCH is 3.8–243×\times× faster in construction, with an average of 9.9×\times× faster. CC suffers from scalability issue in the Reset step in each round (we will discuss more in Sec. 6.3 and Sec. 6.4). As a result, it achieves the best running time with 8 threads, which is close to the performance on sequential algorithms. In general, SPoCH is 10.7–66.1×\times× faster than CC, with an average of 31.5×\times× faster. As shown in Fig. 7, even only considering the time for scoring and contracting, CC is still not competitive to OSRM and is much slower than SPoCH. In general, SPoCH is at least 3.8×\times× faster than all baselines on all graphs. On average across all graphs, SPoCH is 9.9×\times× faster than OSRM, 31.5×\times× faster than CC, 32.4×\times× faster than (sequential) PHAST, and 53.8×\times× than (sequential) RK.

We then analyze the query performance. As mentioned, the quality for a CH generated by a parallel algorithm is generally expected to be lower than those generated sequentially. The results indicate that our algorithm still achieves competitive quality to the best sequential implementation. In general, RK achieves the best query performance in both time and number of vertices visited. SPoCH is competitive to RK, with query times in 1.17–1.79×\times×, and is better than all other baselines in query time, even including the sequential algorithm PHAST. On average, our query time is 1.31×\times× slower than RK, but is 1.25×\times× faster than PHAST, 1.35×\times× faster than OSRM, and 2.07×\times× faster than CC. This indicates the CH generated by our parallel algorithm has close quality to a highly-optimized sequential algorithm, and is better than other parallel versions.

Refer to caption
Figure 5. Heatmap of number of CH edges, query time, and query iteration. All numbers are normalized to ours. Bluer or smaller is better.

Number of CH Edges, Query Time, and Query Iterations.  To compare the number of CH edges, query time, and query iterations of our algorithm against the baselines (OSRM (Luxen and Vetter, 2011), CC (Bühler and Nusser, 2024), RK (Dibbelt et al., 2016), PHAST (Delling et al., 2013b)), we present a heatmap in Fig. 5. Here “query iterations” refers to the average number of vertices visited in an s𝑠sitalic_s-t𝑡titalic_t query, which is machine-independent and roughly indicates the query cost. The numbers in the heatmap are normalized to those of our algorithm. For CH edges, the differences between the baselines and our algorithm are all within 20%, indicating that the space required to store the output CH is similar. For query time and iterations, RK and PHAST, being sequential and following a stricter contraction order, perform slightly better than the parallel implementations. In the worst case, our algorithm is only about twice as slow as the fastest query time. However, since queries finish in microseconds (10−6superscript10610^{-6}10 start_POSTSUPERSCRIPT - 6 end_POSTSUPERSCRIPT seconds), this twofold slowdown is negligible. Among all parallel implementations, our algorithm achieves the best average query time and iterations. In summary, our algorithm offers significantly faster construction performance while maintaining competitive output graph size and query time.

Refer to caption
Figure 6. Self-relative speedups of our algorithm on the three graph categories. Higher is better.

6.3. Scalability

Self-Relative Speedup.  We first show the self-relative speedup of our algorithms on all the graphs in Fig. 6. We vary the number of processors from 1111 to 96969696h (192192192192 hyperthreads). Our self-relative speedup is 29.129.129.129.1–61.7×61.7\times61.7 × on road graphs, 37.937.937.937.9–70.8×70.8\times70.8 × on synthetic graphs and 27.927.927.927.9–67.1×67.1\times67.1 × on k𝑘kitalic_k-NN graphs. This shows that our algorithm achieves high parallelism on all tested graphs.

Scalability Breakdown.  To validate our claims of improving parallelism in CH preprocessing and to illustrate the limitations of existing state-of-the-art parallel implementations, we test the running times of three parallel algorithms—SPoCH, CC, and OSRM—with cores varying from 1111 to 96969696h (192192192192 hyperthreads) in Fig. 7. We select three graphs: AF, CHN7, and GL5, each from a different category. As mentioned, all parallel implementations roughly follow Vetter’s algorithm, so we split the total time into three parts: Score (running WPSes and recomputing vertex scores), Contracting (finding an independent set, contracting them and adding shortcuts), and Others.

Refer to caption
Figure 7. Running time of SPoCH (Ours), OSRM and CH-Constructor (CC) across different core counts. Lower is better. The fastest running time across all core counts is given above its bar. “96h”: 96 cores with hyper-threads. As the “Others” cost in CC dominates in most cases, we also report a version of CC that excludes this time, referred to as CC’.

CC does not achieve satisfactory scalability, and achieves its best performance with 8 threads. The major issue lies in a Reset step that clears arrays at the beginning of each round. This cost increases drastically with the number of threads. To clarify the breakdown between Score and Contract, we also draw a figure for CC without the Others part, referred to as CC′superscriptCC′\textsf{CC}^{\prime}{}CC start_POSTSUPERSCRIPT ′ end_POSTSUPERSCRIPT. The sequential running time for the three algorithms are close. However, the running time for both CC and OSRM flattens after 16 threads. Especially for CC, even without considering the resetting time in the Others part, its time in Contract remains the same or even increases when more than 16 threads are used. This is mainly due to the use of a lock-based structure when adding shortcuts, since using more threads may result in higher contention and further degrade the performance. Our solution using parallel hash table effectively avoids this issue, and remains scalable until using all 192 (hyper)threads.

Refer to caption
Refer to caption
Figure 8. Per-round breakdown of SPoCH (Ours), OSRM and CH-Constructor (CC) on four graphs. The x-axis shows the round number. The top row shows the number of contracted vertices (blue circles), the number of WPSes (orange squares), and the number of remaining overlay vertices (green diamonds) in each round. The bottom row shows the accumulated time for local search (dark blue), scoring (light blue), contracting (coral), and others (red).

OSRM has reasonable scalability up to 16 threads, and its running time remains the same when more threads are used. One possible reason lies in the parallel granularity for computing WPS—OSRM employs parallelism on the level of all vertices in VAsubscript𝑉𝐴\mathit{V_{A}}italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT (vertices that require score recomputing). In particular, all vertices in VAsubscript𝑉𝐴\mathit{V_{A}}italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT are parallelized, but each vertex v∈VA𝑣subscript𝑉𝐴v\in\mathit{V_{A}}italic_v ∈ italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT itself is processed sequentially, which includes a total of Nin⁢(v)subscript𝑁in𝑣N_{\scriptsize\mbox{\it in}}(v)italic_N start_POSTSUBSCRIPT in end_POSTSUBSCRIPT ( italic_v ) executions of Dijkstra’s algorithm. In later rounds when most vertices have been contracted, there are only a few vertices in VAsubscript𝑉𝐴\mathit{V_{A}}italic_V start_POSTSUBSCRIPT italic_A end_POSTSUBSCRIPT, but their degree may have become large, causing a huge amount of work executed sequentially. Therefore, its poor performance is due to insufficient parallelism in later rounds. This issue is also reflected in Fig. 8, which we discuss in Sec. 6.4. In this case, the WPSes are close to a sequential execution, and do not benefit from having more threads. Our solution that batches all WPS sources and parallelizes them as a whole effectively avoids this issue, and can efficiently utilize more threads.

On the three tested graphs, SPoCH shows high scalability up to 96h, and always has better performance with more threads. The self-relative speedup is 53.3×53.3\times53.3 × on AF, 37.9×37.9\times37.9 × on CHN7, and 56.3×56.3\times56.3 × on GL5. All techniques in SPoCH are carefully optimized for high parallelism, such as using lock-free data structures and batching WPSes to run in parallel. Therefore, with a reasonable sequential cost, the good scalability guarantees low parallel running time.

6.4. In-Depth Study for Parallel Algorithms

Recall that parallel CH construction algorithms select an independent set in each round, contract them all in parallel, and repeat until the graph becomes empty. Therefore, the performance and statistics of each round may help to better understand the process of parallel algorithms. In this section, we present an in-depth study for the performance breakdown in each round, for all parallel implementations SPoCH, CC and OSRM. We present the results on four graphs in Fig. 8, which includes the number of vertices processed and accumulated running time up to each round.

Number of Vertices Processed.  In Fig. 8, we present the accumulated numbers of vertices left in the overlay graph (|VO|subscript𝑉O|\mathit{V_{\scriptsize\mbox{\it O}}}|| italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT |), the number of vertices contracted (VFsubscript𝑉𝐹\mathit{V_{F}}italic_V start_POSTSUBSCRIPT italic_F end_POSTSUBSCRIPT), as well as the number of WPSes (i.e., the number of executions of Dijkstra’s algorithm). We first note that OSRM incurs much more rounds than SPoCH and CC—up to 4×\times× more rounds than SPoCH. The larger number of rounds for OSRM is caused by how the independent set is selected. OSRM selects a vertex when it has the minimum score in its 2-hop neighborhood, while SPoCH uses 1-hop neighborhood. The goal of using 2-hop neighborhood is to avoid using complicated concurrent data structure to deal with shortcut insertion, but it results in a much slower contraction process than SPoCH. On the contrary, SPoCH uses the efficient data structure introduced in Sec. 4, which allows contracting all vertices with minimum score in their 1-hop neighborhood, leading to fewer rounds.

For all three algorithms, the size of the overlay graph decreases quickly. On CHN8, the trend of |VO|subscript𝑉O|\mathit{V_{\scriptsize\mbox{\it O}}}|| italic_V start_POSTSUBSCRIPT O end_POSTSUBSCRIPT | decreases linearly on the log scale. Indeed on a chain, since almost all vertices have the same edge difference, the contraction is mostly determined by the random priority. In this case, a constant fraction of vertices can be selected in expectation (Shun et al., 2015). On other graphs, the size of the overlay graph decreases rapidly in the first few rounds, and then slows down. In general, for all implementations, the overlay graph size decreases by more than 99% in the first 10–20 rounds. This implies great potential of parallelism in the construction of CH.

We then compare the number of WPSes executed for each algorithm. As mentioned, one of our efforts is to reduce the number of WPSes performed to optimize the overall performance. SPoCH incurs significantly fewer Witness Path Searches (WPSes) compared to CC and OSRM. In each round, especially in later rounds, SPoCH requires far fewer WPSes than the total number of vertices in the overlay graph. This reduction in the number of WPS demonstrates the efficiency of our approach in limiting unnecessary searches.

Performance Breakdown.  We now analyze the accumulated running time across rounds. In the first 10%percent1010\%10 % of the rounds, the running time for both algorithms grows quickly. As mentioned, more than 99% of the vertices may be contracted just in the first 10–20 rounds. As the algorithm proceeds, the overlay graph becomes much smaller. On the four graphs, all algorithms reach a small overlay graph with 105superscript10510^{5}10 start_POSTSUPERSCRIPT 5 end_POSTSUPERSCRIPT within 30 rounds, after which very little computation is required. Indeed, our algorithm spends the majority of its time in the first several rounds. As the size of the overlay graph decreases, the running time for the last 90% of rounds is small. Indeed, all the steps become cheap after a few rounds. This indicates the effectiveness of our solution to maintain the overlay graph—when the overlay graph becomes small, the maintenance cost also shrinks proportionally. This benefit comes from our technique of maintaining the overlay graph in a lazy manner. In later rounds where only a few shortcuts are generated in each round, SPoCH delays the process of combining them into the CSR until it collects a sufficient number of shortcuts. Thus, after 10101010–20202020 rounds, there are likely only one or two global merges incurred, leading to SPoCH’s high overall performance.

In contrast, the running time for OSRM grows steadily across all rounds. This indicates that even processing a small overlay graph, OSRM may spend a noticeable amount of time. As we discussed in Sec. 6.3, this is likely due to insufficient parallelism in later rounds. For CC, as mentioned, the use of inefficient data structures to maintain shortcuts dramatically increases its running time in the Contract step, leading to overall unsatisfactory performance.

In summary, our design allows for contracting a large fraction of vertices in each round, leading to a very fast contraction process. Our new design for the LocalSearch step also reduces the number for WPSes performed. Our data structure also avoids a high cost to maintain a small overlay graph. As a result, SPoCH outperforms both parallel baselines due to new designs in both algorithm.

Refer to caption
Figure 9. Running time with and without the Lazy Combination Optimization. For each graph, the left bar corresponds to the cost breakdown without lazy combination, while the right bar represent the breakdown with the optimization. The cost of combination is counted in the Contract step.
Graph Road Synthetic k𝑘kitalic_k-NN
CA AO AF NA AS EU TRCE BUB CHN7 CHN8 HT5 CH2 CH5 GL2 GL5 GL10
Build ParlayLib (Blelloch et al., 2020a) 1.00 2.25 7.32 23.1 23.5 32.9 25.2 34.7 1.77 14.9 0.68 0.87 115 2.95 9.78 41.6
OpenCilk (OpenCilk, 2020) 0.97 2.21 7.37 23.1 23.7 32.5 24.9 32.4 1.82 15.4 0.67 0.86 111 2.96 9.71 41.2
TBB (Intel Threading Building Blocks, [n. d.]) 1.03 2.33 7.66 23.5 24.0 32.8 25.1 34.3 1.93 15.9 0.50 0.96 114 3.30 10.3 42.3
Query ParlayLib 7.99 27.9 17.2 93.3 60.4 137 376 452 5.59 7.74 6.16 3.55 244 1.93 10.2 33.9
OpenCilk 7.91 27.8 17.1 93.3 60.7 141 379 454 5.56 7.78 6.24 3.53 239 1.93 10.4 32.8
TBB 7.90 27.9 17.3 94.4 57.6 136 377 452 5.64 7.76 6.27 3.54 242 1.92 10.3 32.5
Table 5. Comparison of build time (in seconds) and query time (in microseconds) for all tested schedulers across all graphs. Smaller is better. For each graph, the best build time and best query time are bolded and underlined.

Evaluating the Lazy Combination of Shortcut Edges.  As mentioned, one of our optimizations is to combine the shortcuts E+superscript𝐸\mathit{E^{+}}italic_E start_POSTSUPERSCRIPT + end_POSTSUPERSCRIPT lazily with the edges of the overlay graph EOsubscript𝐸O{\mathit{E_{\scriptsize\mbox{\it O}}}}italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT. This is to reduce and amortize the cost of updating the CSR for EOsubscript𝐸O{\mathit{E_{\scriptsize\mbox{\it O}}}}italic_E start_POSTSUBSCRIPT O end_POSTSUBSCRIPT. Fig. 9 shows the running time with and without this optimization. For each graph, the left bar represents the breakdown that combines in each round (without lazy combination), while the right bar represents that with the optimization. The cost is counted in the Contract step. By combining the shortcuts lazily instead of doing it every round, the cost of the Contract step is reduced from 39.9% to 19.8% of the overall time on average. Across all graphs, this optimization improves the performance by 1.1–2.7×\times× for the total running time.

Impact of Different Schedulers.  To evaluate the impact of different schedulers to the performance of SPoCH, we conduct an experiment comparing ParlayLib (Blelloch et al., 2020a), OpenCilk (OpenCilk, 2020), and TBB (Intel Threading Building Blocks, [n. d.]). Our algorithmic framework remains the same across all tests, ensuring that any observed performance differences stem primarily from the underlying scheduler rather than the core algorithm itself. Across all schedulers, the average build time is within 10%percent1010\%10 % of the fastest result, and the average query time is within 2%percent22\%2 %. Hence, the choice of scheduler has a negligible effect on the performance of SPoCH. We present detailed running times in Tab. 5.

7. Related Work

Computing the shortest paths on a graph is one of the most well-studied problems in computer science. We refer the audience to the excellent surveys on this topic, including those by Bast et al. (Bast et al., 2016), Madkoure et al. (Madkour et al., 2017), and Sommer (Sommer, 2014). These surveys all discuss contraction hierarchies (CH) in detail. Meanwhile, we note that the main applications of the CH are on sparse networks such as roadmaps (transition networks). We acknowledge that many interesting algorithms have been proposed for relevant but different problems, such as on social networks, queries among all pairs or between a batch of sources and destinations, on dynamic graphs, and time-dependant queries. A brief list of recent work on computing shortest paths includes (Zhang et al., 2022, 2021; Qiu et al., 2022; Li et al., 2020a; Yu et al., 2020; Li et al., 2020b; Zhang and Yu, 2020; Huang et al., 2021; Gong et al., 2024).

The idea of CH was proposed by Geisberger et al. (Geisberger et al., 2008), based on simplifying highway hierarchies (Sanders and Schultes, 2006; Knopp et al., 2007) and highway node routing (Schultes and Sanders, 2007). CH has achieved notable success in practical applications, and fostered numerous later studies on relevant problems. Examples include the time-dependent versions (Batz et al., 2010, 2013), parallel and distributed versions (Kieritz et al., 2010; Vetter, 2009), on dynamic graphs (Ouyang et al., 2020), and more algorithmic optimizations (Geisberger et al., 2012; Delling et al., 2017; Hespe and Sanders, 2019). CH is also used to parallelize SSSP queries, such as in PHAST (Group, 2024; Delling et al., 2013b). Another stream of research is to derive theoretical guarantees for CH (Dibbelt et al., 2016; Blum et al., 2021; Funke and Storandt, 2015; Abraham et al., 2010; Hamann and Strasser, 2018; Columbus and Bauer, 2009). These analyses are mostly parameterized, based on some graph invariants such as tree depths, tree widths, and diameters.

There have been many studies on parallelizing CH construction and relevant techniques. Vetter’s work (Vetter, 2009) is the earliest and inspired many of the later studies (Luxen and Vetter, 2024, 2011; Chen et al., 2024; Karimi et al., 2019, 2020; Kieritz et al., 2010; Bühler and Nusser, 2024). We reviewed Vetter’s approach in Sec. 2.2. OSRM by Luxen and Vetter (Luxen and Vetter, 2024, 2011) implements Vetter’s approach, and it is considered the SOTA parallel CH construction. CH-Constructor (Bühler and Nusser, 2024) is another open-source software for parallel CH construction. We compared to OSRM and CH-Constructor since they have open-source code available. The others focus on different settings, such as distributed (Kieritz et al., 2010), GPU (Karimi et al., 2019, 2020), and on edge contraction (Chen et al., 2024). Among them, an existing GPU algorithm (Karimi et al., 2020) also prune unnecessary shortcuts. SPoCH has the same motivation for pruning, but uses a different methodology by leveraging memoization to further reduce the cost. We believe that some of our algorithmic techniques, such as LocalSearch with memoization, are applicable to other settings such as GPUs. However, we acknowledge that part of the implementation, such as the parallel data structures, will require careful redesign, which we leave as future work.

We note that there are other shortest-path algorithms that can provide different construction-query trade-offs or are designed for other graph types. Some of these algorithms include transit node routing (Bast et al., 2007), hub labeling (Abraham et al., 2011, 2012), pruned landmark labeling (Akiba et al., 2013), highway labeling (Akiba et al., 2014), and ALT (Goldberg and Harrelson, 2005). Indeed, many of these approaches (Arz et al., 2013; Bast et al., 2006, 2007; Abraham et al., 2011, 2012; Delling et al., 2013a; Bauer et al., 2010; Delling et al., 2013b) use CH as a subroutine. We believe that the faster CH construction presented in this paper can improve such trade-offs, and we leave it as future work.

8. Conclusions and Future Work

In this paper, we propose SPoCH (Scalable Parallelization of Contraction Hierarchies), a parallel algorithm for constructing contraction hierarchies. Our key insights include algorithm redesign to introduce the LocalSearch step, which allows for batching, memoization and pruning, as well as leveraging parallel data structures. In this way, SPoCH effectively reduces the total work and enhances parallelism. Across 16 graphs of varying sizes and average degrees (including road networks, synthetic graphs, and k𝑘kitalic_k-NN graphs), SPoCH consistently outperforms four other SOTA sequential and parallel baselines by 3.83–243×\times×, while maintaining competitive query performance. On a 96-core machine, SPoCH delivers self-relative speedups of 27.9–70.8×\times×. We conduct in-depth experiments to analyze the improvements of our techniques.

An open question remains as to whether parallel CHs can handle graphs with high degrees (e.g., social networks). Some interesting future directions include applying our new CH construction algorithm to other distance queries, such as single-source shortest paths, all-pairs shortest paths, or distance oracles.

Acknowledgements.
This work is supported by NSF grants CCF-2103483, IIS-2227669, and TI-2346223, NSF CAREER Awards CCF-2238358 and CCF-2339310, the UCR Regents Faculty Development Award, and the Google Research Scholar Program. We thank the anonymous reviewers for their useful comments.

References

  • (1)
  • Abeywickrama et al. (2016) Tenindra Abeywickrama, Muhammad Aamir Cheema, and David Taniar. 2016. k-Nearest Neighbors on Road Networks: A Journey in Experimentation and In-Memory Implementation. Proceedings of the VLDB Endowment (PVLDB) 9, 6 (2016).
  • Abraham et al. (2011) Ittai Abraham, Daniel Delling, Andrew V Goldberg, and Renato F Werneck. 2011. A hub-based labeling algorithm for shortest paths in road networks. In International Symposium on Experimental Algorithms (SEA). Springer, 230–241.
  • Abraham et al. (2012) Ittai Abraham, Daniel Delling, Andrew V Goldberg, and Renato F Werneck. 2012. Hierarchical hub labelings for shortest paths. In European Symposium on Algorithms (ESA). Springer, 24–35.
  • Abraham et al. (2010) Ittai Abraham, Amos Fiat, Andrew V Goldberg, and Renato F Werneck. 2010. Highway dimension, shortest paths, and provably efficient algorithms. In ACM-SIAM Symposium on Discrete Algorithms (SODA). SIAM, 782–793.
  • Akiba et al. (2014) Takuya Akiba, Yoichi Iwata, Ken-ichi Kawarabayashi, and Yuki Kawata. 2014. Fast shortest-path distance queries on road networks by pruned highway labeling. In Algorithm Engineering and Experiments (ALENEX). SIAM, 147–154.
  • Akiba et al. (2013) Takuya Akiba, Yoichi Iwata, and Yuichi Yoshida. 2013. Fast exact shortest-path distance queries on large networks by pruned landmark labeling. In ACM SIGMOD International Conference on Management of Data (SIGMOD). 349–360.
  • Arora et al. (2001) Nimar S Arora, Robert D Blumofe, and C Greg Plaxton. 2001. Thread scheduling for multiprogrammed multiprocessors. Theory of Computing Systems (TOCS) 34, 2 (2001), 115–144.
  • Arz et al. (2013) Julian Arz, Dennis Luxen, and Peter Sanders. 2013. Transit node routing reconsidered. In International symposium on experimental algorithms. Springer, 55–66.
  • Bast et al. (2016) Hannah Bast, Daniel Delling, Andrew Goldberg, Matthias Mü ller Hannemann, Thomas Pajor, Peter Sanders, Dorothea Wagner, and Renato F Werneck. 2016. Route planning in transportation networks. Algorithm engineering: Selected results and surveys (2016), 19–80.
  • Bast et al. (2006) Hannah Bast, Stefan Funke, and Domagoj Matijevic. 2006. Ultrafast Shortest-Path Queries via Transit Nodes. The Shortest Path Problem 74 (2006), 175–192.
  • Bast et al. (2007) Holger Bast, Stefan Funke, Peter Sanders, and Dominik Schultes. 2007. Fast routing in road networks with transit nodes. Science 316, 5824 (2007), 566–566.
  • Batz et al. (2010) Gernot Veit Batz, Robert Geisberger, Sabine Neubauer, and Peter Sanders. 2010. Time-dependent contraction hierarchies and approximation. In International Symposium on Experimental Algorithms (SEA). Springer, 166–177.
  • Batz et al. (2013) G Veit Batz, Robert Geisberger, Peter Sanders, and Christian Vetter. 2013. Minimum time-dependent travel times with contraction hierarchies. Journal of Experimental Algorithmics (JEA) 18 (2013), 1–1.
  • Bauer et al. (2010) Reinhard Bauer, Daniel Delling, Peter Sanders, Dennis Schieferdecker, Dominik Schultes, and Dorothea Wagner. 2010. Combining hierarchical and goal-directed speed-up techniques for dijkstra’s algorithm. Journal of Experimental Algorithmics (JEA) 15 (2010), 2–1.
  • Ben-David et al. (2016) Naama Ben-David, Guy E. Blelloch, Jeremy T. Fineman, Phillip B. Gibbons, Yan Gu, Charles McGuffey, and Julian Shun. 2016. Parallel Algorithms for Asymmetric Read-Write Costs. In ACM Symposium on Parallelism in Algorithms and Architectures (SPAA).
  • Blelloch et al. (2020a) Guy E. Blelloch, Daniel Anderson, and Laxman Dhulipala. 2020a. ParlayLib — a toolkit for parallel algorithms on shared-memory multicore machines. In ACM Symposium on Parallelism in Algorithms and Architectures (SPAA). 507–509.
  • Blelloch and Dobson (2022) Guy E Blelloch and Magdalen Dobson. 2022. Parallel Nearest Neighbors in Low Dimensions with Batch Updates. In Algorithm Engineering and Experiments (ALENEX). SIAM, 195–208.
  • Blelloch et al. (2020b) Guy E. Blelloch, Jeremy T. Fineman, Yan Gu, and Yihan Sun. 2020b. Optimal parallel algorithms in the binary-forking model. In ACM Symposium on Parallelism in Algorithms and Architectures (SPAA). 89–102.
  • Blelloch et al. (2010) Guy E. Blelloch, Phillip B. Gibbons, and Harsha Vardhan Simhadri. 2010. Low depth cache-oblivious algorithms. In ACM Symposium on Parallelism in Algorithms and Architectures (SPAA).
  • Blum et al. (2021) Johannes Blum, Stefan Funke, and Sabine Storandt. 2021. Sublinear search spaces for shortest path planning in grid and road networks. Journal of Combinatorial Optimization 42, 2 (2021), 231–257.
  • Blumofe and Leiserson (1999) Robert D. Blumofe and Charles E. Leiserson. 1999. Scheduling multithreaded computations by work stealing. J. ACM 46, 5 (1999), 720–748.
  • Bühler and Nusser (2024) Stefan Bühler and André Nusser. 2024. Contraction Hierarchies Constructor. https://doi.org/10.5281/zenodo.14008202
  • Chen et al. (2024) Zi Chen, Xinyu Ji, Long Yuan, Xuemin Lin, Wenjie Zhang, and Shan Huang. 2024. Parallel Contraction Hierarchies Construction on Road Networks. IEEE Transactions on Knowledge and Data Engineering (2024).
  • Columbus and Bauer (2009) Tobias Columbus and Reinhard Bauer. 2009. On the complexity of contraction hierarchies. Student Thesis, Karlsruhe Inst. Technol., Karlsruhe, Germany (2009).
  • contributors (2024) RoutingKit contributors. 2024. RoutingKit. https://github.com/RoutingKit/RoutingKit.
  • Delling et al. (2013b) Daniel Delling, Andrew V Goldberg, Andreas Nowatzyk, and Renato F Werneck. 2013b. PHAST: Hardware-accelerated shortest path trees. J. Parallel and Distrib. Comput. 73, 7 (2013), 940–952.
  • Delling et al. (2017) Daniel Delling, Andrew V Goldberg, Thomas Pajor, and Renato F Werneck. 2017. Customizable route planning in road networks. Transportation Science 51, 2 (2017), 566–591.
  • Delling et al. (2013a) Daniel Delling, Andrew V Goldberg, and Renato F Werneck. 2013a. Hub label compression. In International symposium on experimental algorithms. Springer, 18–29.
  • Dhulipala et al. (2021) Laxman Dhulipala, Guy E. Blelloch, and Julian Shun. 2021. Theoretically efficient parallel graph algorithms can be fast and scalable. ACM Transactions on Parallel Computing (TOPC) 8, 1 (2021), 1–70.
  • Dhulipala et al. (2024) Laxman Dhulipala, Xiaojun Dong, Kishen Gowda, and Yan Gu. 2024. Optimal Parallel Algorithms for Dendrogram Computation and Single-Linkage Clustering. In ACM Symposium on Parallelism in Algorithms and Architectures (SPAA).
  • Dibbelt et al. (2016) Julian Dibbelt, Ben Strasser, and Dorothea Wagner. 2016. Customizable contraction hierarchies. J. Experimental Algorithmics 21 (2016), 1–49.
  • Dijkstra (1959) Edsger W. Dijkstra. 1959. A note on two problems in connexion with graphs. Numerische mathematik 1, 1 (1959).
  • Dong et al. (2024a) Xiaojun Dong, Laxman Dhulipala, Yan Gu, and Yihan Sun. 2024a. Parallel Integer Sort: Theory and Practice. In ACM Symposium on Principles and Practice of Parallel Programming (PPOPP).
  • Dong et al. (2024b) Xiaojun Dong, Yan Gu, Yihan Sun, and Letong Wang. 2024b. Brief Announcement: PASGAL: Parallel And Scalable Graph Algorithm Library. In ACM Symposium on Parallelism in Algorithms and Architectures (SPAA).
  • Dong et al. (2021) Xiaojun Dong, Yan Gu, Yihan Sun, and Yunming Zhang. 2021. Efficient Stepping Algorithms and Implementations for Parallel Shortest Paths. In ACM Symposium on Parallelism in Algorithms and Architectures (SPAA). 184–197.
  • Dong et al. (2023) Xiaojun Dong, Letong Wang, Yan Gu, and Yihan Sun. 2023. Provably Fast and Space-Efficient Parallel Biconnectivity. In ACM Symposium on Principles and Practice of Parallel Programming (PPOPP). 52–65.
  • Fonollosa et al. (2015) Jordi Fonollosa, Sadique Sheik, Ramón Huerta, and Santiago Marco. 2015. Reservoir computing compensates slow response of chemosensor arrays exposed to fast varying gas concentrations in continuous monitoring. Sensors and Actuators B: Chemical 215 (2015), 618–629.
  • Funke et al. (2015) Stefan Funke, André Nusser, and Sabine Storandt. 2015. Placement of loading stations for electric vehicles: No detours necessary! Journal of Artificial Intelligence Research 53 (2015), 633–658.
  • Funke et al. (2019) Stefan Funke, Tobias Rupp, André Nusser, and Sabine Storandt. 2019. PATHFINDER: storage and indexing of massive trajectory sets. In Proceedings of the 16th International Symposium on Spatial and Temporal Databases. 90–99.
  • Funke and Storandt (2015) Stefan Funke and Sabine Storandt. 2015. Provable efficiency of contraction hierarchies with randomized preprocessing. In International Symposium on Algorithms and Computation. Springer, 479–490.
  • Geisberger et al. (2008) Robert Geisberger, Peter Sanders, Dominik Schultes, and Daniel Delling. 2008. Contraction hierarchies: Faster and simpler hierarchical routing in road networks. In International Workshop on Experimental Algorithms (WEA). Springer, 319–333.
  • Geisberger et al. (2012) Robert Geisberger, Peter Sanders, Dominik Schultes, and Christian Vetter. 2012. Exact routing in large road networks using contraction hierarchies. Transportation Science 46, 3 (2012), 388–404.
  • Goldberg and Harrelson (2005) Andrew V Goldberg and Chris Harrelson. 2005. Computing the shortest path: A search meets graph theory. In ACM-SIAM Symposium on Discrete Algorithms (SODA), Vol. 5. 156–165.
  • Gong et al. (2024) Zengyang Gong, Yuxiang Zeng, and Lei Chen. 2024. Querying shortest path on large time-dependent road networks with shortcuts. In IEEE International Conference on Data Engineering (ICDE). IEEE, 4532–4544.
  • Group (2024) KIT ITI Algorithmics Group. 2024. PHAST. https://github.com/kit-algo/ULTRA-PHAST.
  • Gu et al. (2023) Yan Gu, Ziyang Men, Zheqi Shen, Yihan Sun, and Zijin Wan. 2023. Parallel Longest Increasing Subsequence and van Emde Boas Trees. In ACM Symposium on Parallelism in Algorithms and Architectures (SPAA).
  • Hamann and Strasser (2018) Michael Hamann and Ben Strasser. 2018. Graph bisection with pareto optimization. J. Experimental Algorithmics 23 (2018), 1–34.
  • Hespe and Sanders (2019) Demian Hespe and Peter Sanders. 2019. More Hierarchy in Route Planning Using Edge Hierarchies. In Symposium on Algorithmic Approaches for Transportation Modelling, Optimization, and Systems (ATMOS).
  • Huang et al. (2021) Shuai Huang, Yong Wang, Tianyu Zhao, and Guoliang Li. 2021. A learning-based method for computing shortest path distances on road networks. In IEEE International Conference on Data Engineering (ICDE). IEEE, 360–371.
  • Huerta et al. (2016) Ramon Huerta, Thiago Mosqueiro, Jordi Fonollosa, Nikolai F Rulkov, and Irene Rodriguez-Lujan. 2016. Online decorrelation of humidity and temperature in chemical sensors for continuous monitoring. Chemometrics and Intelligent Laboratory Systems 157 (2016), 169–176.
  • Intel Threading Building Blocks ([n. d.]) Intel Threading Building Blocks [n. d.]. Intel Threading Building Blocks (TBB). https://www.threadingbuildingblocks.org.
  • Karimi et al. (2019) Roozbeh Karimi, David M Koppelman, and Chris J Michael. 2019. GPU road network graph contraction and SSSP query. In Proceedings of the ACM International Conference on Supercomputing. 250–260.
  • Karimi et al. (2020) Roozbeh Karimi, David M Koppelman, and Chris J Michael. 2020. Fast GPU Graph Contraction by Combining Efficient Shallow Searches and Post-Culling. In IEEE Conference on High Performance Extreme Computing (HPEC). IEEE, 1–7.
  • Kieritz et al. (2010) Tim Kieritz, Dennis Luxen, Peter Sanders, and Christian Vetter. 2010. Distributed time-dependent contraction hierarchies. In International Symposium on Experimental Algorithms (SEA). Springer, 83–93.
  • Knopp et al. (2007) Sebastian Knopp, Peter Sanders, Dominik Schultes, Frank Schulz, and Dorothea Wagner. 2007. Computing many-to-many shortest paths using highway hierarchies. In Algorithm Engineering and Experiments (ALENEX). SIAM, 36–45.
  • Li et al. (2020a) Lingxiao Li, Muhammad Aamir Cheema, Mohammed Eunus Ali, Hua Lu, and David Taniar. 2020a. Continuously monitoring alternative shortest paths on road networks. In Proceedings of the VLDB Endowment (PVLDB). Association for Computing Machinery, 2243–2255.
  • Li et al. (2020b) Lei Li, Mengxuan Zhang, Wen Hua, and Xiaofang Zhou. 2020b. Fast query decomposition for batch shortest path processing in road networks. In IEEE International Conference on Data Engineering (ICDE). IEEE, 1189–1200.
  • Liu et al. (2025) Youzhe Liu, Xiaojun Dong, Yan Gu, and Yihan Sun. 2025. Parallel k𝑘kitalic_k-Core Decomposition: Theory and Practice. ACM SIGMOD International Conference on Management of Data (SIGMOD) 3, 3 (2025).
  • Luxen and Vetter (2011) Dennis Luxen and Christian Vetter. 2011. Real-time routing with OpenStreetMap data. In SIGSPATIAL international conference on advances in geographic information systems. 513–516.
  • Luxen and Vetter (2024) Dennis Luxen and Christian Vetter. 2024. Open Source Routing Machine. https://github.com/Project-OSRM/osrm-backend.
  • Madkour et al. (2017) Amgad Madkour, Walid G Aref, Faizan Ur Rehman, Mohamed Abdur Rahman, and Saleh Basalamah. 2017. A survey of shortest-path algorithms. arXiv preprint:1705.02044 (2017).
  • Meyer and Sanders (2003) Ulrich Meyer and Peter Sanders. 2003. ΔΔ\Deltaroman_Δ-stepping: a parallelizable shortest path algorithm. Journal of Algorithms 49, 1 (2003), 114–152.
  • OpenCilk (2020) OpenCilk 2020. OpenCilk. https://www.opencilk.org/.
  • OpenStreetMap contributors (2010) OpenStreetMap contributors. 2010. OpenStreetMap. https://www.openstreetmap.org/.
  • Ouyang et al. (2020) Dian Ouyang, Long Yuan, Lu Qin, Lijun Chang, Ying Zhang, and Xuemin Lin. 2020. Efficient shortest path index maintenance on dynamic road networks with theoretical guarantees. Proceedings of the VLDB Endowment (PVLDB) 13, 5 (2020), 602–615.
  • Qiu et al. (2022) Yu-Xuan Qiu, Dong Wen, Lu Qin, Wentao Li, Rong-Hua Li, Zhang Ying, et al. 2022. Efficient shortest path counting on large road networks. Proceedings of the VLDB Endowment (PVLDB) (2022).
  • Sanchez et al. (2022) Naroa Coretti Sanchez, Iñigo Martinez, Luis Alonso Pastor, and Kent Larson. 2022. On the simulation of shared autonomous micro-mobility. Communications in Transportation Research 2 (2022), 100065.
  • Sanders and Schultes (2006) Peter Sanders and Dominik Schultes. 2006. Engineering highway hierarchies. In European Symposium on Algorithms. Springer, 804–816.
  • Sanders et al. (2014) Peter Sanders, Christian Schulz, and Dorothea Wagner. 2014. Benchmarking for graph clustering and partitioning. Encyclopedia of social network analysis and mining Springer (2014).
  • Schultes and Sanders (2007) Dominik Schultes and Peter Sanders. 2007. Dynamic highway-node routing. In Workshop on Experimental and Efficient Algorithms (WEA). Springer, 66–79.
  • Shen et al. (2022) Zheqi Shen, Zijin Wan, Yan Gu, and Yihan Sun. 2022. Many Sequential Iterative Algorithms Can Be Parallel and (Nearly) Work-efficient. In ACM Symposium on Parallelism in Algorithms and Architectures (SPAA).
  • Shun and Blelloch (2014) Julian Shun and Guy E Blelloch. 2014. Phase-concurrent hash tables for determinism. In ACM Symposium on Parallelism in Algorithms and Architectures (SPAA). 96–107.
  • Shun et al. (2015) Julian Shun, Yan Gu, Guy E. Blelloch, Jeremy T. Fineman, and Phillip B Gibbons. 2015. Sequential random permutation, list contraction and tree contraction are highly parallel. In ACM-SIAM Symposium on Discrete Algorithms (SODA). 431–448.
  • Sommer (2014) Christian Sommer. 2014. Shortest-path queries in static networks. ACM Computing Surveys (CSUR) 46, 4 (2014), 1–31.
  • Tinney and Walker (1967) William F Tinney and John W Walker. 1967. Direct solutions of sparse network equations by optimally ordered triangular factorization. Proc. IEEE 55, 11 (1967), 1801–1809.
  • Vetter (2009) Christian Vetter. 2009. Parallel time-dependent contraction hierarchies. Student Research Project (2009), 134.
  • Wan et al. (2024) Zijin Wan, Xiaojun Dong, Letong Wang, Enzuo Zhu, Yan Gu, and Yihan Sun. 2024. Implementations for Parallel Contraction Hierarchies. https://github.com/ucrparlay/Parallel-Contraction-Hierarchy.
  • Wang et al. (2023) Letong Wang, Xiaojun Dong, Yan Gu, and Yihan Sun. 2023. Parallel Strong Connectivity Based on Faster Reachability. ACM SIGMOD International Conference on Management of Data (SIGMOD) 1, 2 (2023), 1–29.
  • Wang et al. (2021) Yiqiu Wang, Shangdi Yu, Laxman Dhulipala, Yan Gu, and Julian Shun. 2021. GeoGraph: A Framework for Graph Processing on Geometric Data. ACM SIGOPS Operating Systems Review 55, 1 (2021), 38–46.
  • Wheatman et al. (2024) Brian Wheatman, Xiaojun Dong, Zheqi Shen, Laxman Dhulipala, Jakub Ł\kacki, Prashant Pandey, and Helen Xu. 2024. BYO: A Unified Framework for Benchmarking Large-Scale Graph Containers. In Proceedings of the VLDB Endowment (PVLDB).
  • Yu et al. (2020) Ziqiang Yu, Xiaohui Yu, Nick Koudas, Yang Liu, Yifan Li, Yueting Chen, and Dingyu Yang. 2020. Distributed processing of k shortest path queries over dynamic road networks. In ACM SIGMOD International Conference on Management of Data (SIGMOD). 665–679.
  • Zhang et al. (2022) Junhua Zhang, Wentao Li, Long Yuan, Lu Qin, Ying Zhang, and Lijun Chang. 2022. Shortest-path queries on complex networks: experiments, analyses, and improvement. Proceedings of the VLDB Endowment (PVLDB) 15, 11 (2022), 2640–2652.
  • Zhang et al. (2021) U Zhang, Long Yuan, Wentao Li, Lu Qin, and Ying Zhang. 2021. Efficient label-constrained shortest path queries on road networks: A tree decomposition approach. Proceedings of the VLDB Endowment (PVLDB) (2021).
  • Zhang and Yu (2020) Yikai Zhang and Jeffrey Xu Yu. 2020. Hub labeling for shortest path counting. In ACM SIGMOD International Conference on Management of Data (SIGMOD). 1813–1828.
  • Zheng et al. (2008) Yu Zheng, Like Liu, Longhao Wang, and Xing Xie. 2008. Learning transportation mode from raw gps data for geographic applications on the web. In International World Wide Web Conference (WWW). 247–256.